from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.filter_df import convert_uf_to_pole
from edsteva.probes.utils.prepare_df import (
    prepare_care_site,
    prepare_note,
    prepare_note_care_site,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_note(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    extra_data: Data,
    length_of_stays: List[float],
    note_types: Union[str, Dict[str, str]],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_note`` algorithm computes $c_(t)$ the availability of clinical documents as follow:

    $$
    c(t) = \frac{n_{note}(t)}{n_{max}}
    $$

    Where $n_{note}(t)$ is the number of clinical documents, $t$ is the month and $n_{max} = \max_{t}(n_{note}(t))$.
    """
    self._metrics = ["c", "n_note"]
    check_tables(data=data, required_tables=["note"])

    note = prepare_note(
        data=data,
        start_date=start_date,
        end_date=end_date,
        note_types=note_types,
    )

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    ).drop(columns=["visit_occurrence_source_value", "date"])

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_relationship=care_site_relationship,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
    )

    note_hospital = get_hospital_note(note, visit_occurrence, care_site)
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    note_predictor_by_level = {hospital_name: note_hospital}

    # UF selection
    if not hospital_only(care_site_levels=care_site_levels):
        if extra_data:  # pragma: no cover
            note_uf, note_uc, note_uh = get_note_detail(
                extra_data,
                note,
                visit_occurrence,
                care_site,
            )
            uf_name = CARE_SITE_LEVEL_NAMES["UF"]
            note_predictor_by_level[uf_name] = note_uf
            uc_name = CARE_SITE_LEVEL_NAMES["UC"]
            note_predictor_by_level[uc_name] = note_uc
            uh_name = CARE_SITE_LEVEL_NAMES["UH"]
            note_predictor_by_level[uh_name] = note_uh

            note_pole = get_pole_note(note_uf, care_site, care_site_relationship)
            pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
            note_predictor_by_level[pole_name] = note_pole
        else:
            logger.info("Note data are only available at hospital level")
            care_site_levels = ["Hospital"]

    # Concatenate all predictors
    note_predictor = concatenate_predictor_by_level(
        predictor_by_level=note_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    if is_koalas(note_predictor):
        note_predictor.spark.cache()

    return compute_completeness(self, note_predictor)


def compute_completeness(
    self,
    note_predictor: DataFrame,
):
    partition_cols = [*self._index.copy(), "date"]

    n_note = (
        note_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"note_id": "nunique"})
        .rename(columns={"note_id": "n_note"})
    )
    n_note = to("pandas", n_note)
    n_note = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_note,
        partition_cols=partition_cols,
    )

    partition_cols = list(set(partition_cols) - {"date"})
    max_note = (
        n_note.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"n_note": "max"})
        .rename(columns={"n_note": "max_n_note"})
    )

    note_predictor = n_note.merge(
        max_note,
        on=partition_cols,
    )

    note_predictor["c"] = note_predictor["max_n_note"].where(
        note_predictor["max_n_note"] == 0,
        note_predictor["n_note"] / note_predictor["max_n_note"],
    )
    return note_predictor.drop(columns="max_n_note")


def get_hospital_note(
    note: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    note_hospital = note.merge(visit_occurrence, on="visit_occurrence_id")
    note_hospital = note_hospital.drop(columns="visit_occurrence_id")
    note_hospital = note_hospital.merge(care_site, on="care_site_id")
    if is_koalas(note_hospital):
        note_hospital = note_hospital.spark.cache()

    return note_hospital


def get_note_detail(
    extra_data: Data,
    note: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):  # pragma: no cover
    note_detail = prepare_note_care_site(extra_data=extra_data, note=note)
    note_detail = note_detail.merge(care_site, on="care_site_id")
    note_detail = note_detail.merge(
        visit_occurrence.drop(columns="care_site_id"), on="visit_occurrence_id"
    ).drop(columns="visit_occurrence_id")

    uf_name = CARE_SITE_LEVEL_NAMES["UF"]
    note_uf = note_detail[note_detail["care_site_level"] == uf_name]
    uc_name = CARE_SITE_LEVEL_NAMES["UC"]
    note_uc = note_detail[note_detail["care_site_level"] == uc_name]
    uh_name = CARE_SITE_LEVEL_NAMES["UH"]
    note_uh = note_detail[note_detail["care_site_level"] == uh_name]
    if is_koalas(note_detail):
        note_uf = note_uf.spark.cache()
        note_uc = note_uc.spark.cache()
        note_uh = note_uh.spark.cache()

    return note_uf, note_uc, note_uh


def get_pole_note(
    note_uf: DataFrame,
    care_site: DataFrame,
    care_site_relationship: DataFrame,
):  # pragma: no cover
    care_site_cols = list(
        set(
            [
                "care_site_short_name",
                "care_site_level",
                "care_site_specialty",
                "specialties_set",
                "care_sites_set",
            ]
        ).intersection(note_uf.columns)
    )
    note_pole = convert_uf_to_pole(
        table=note_uf.drop(columns=care_site_cols),
        table_name="note_uf",
        care_site_relationship=care_site_relationship,
    )

    note_pole = note_pole.merge(care_site, on="care_site_id")

    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    note_pole = note_pole[note_pole["care_site_level"] == pole_name]

    if is_koalas(note_pole):
        note_pole = note_pole.spark.cache()

    return note_pole



"""
The default tables loaded when instanciating a [``HiveData``][edsteva.io.hive.HiveData]
or a [``PostgresData``][edsteva.io.postgres.PostgresData]
"""
default_tables_to_save = [
    "person",
    "visit_occurrence",
    "visit_detail",
    "condition_occurrence",
    "procedure_occurrence",
    "care_site",
    "concept",
]

"""
The default columns loaded when instanciating a [``HiveData``][edsteva.io.hive.HiveData]
or a [``PostgresData``][edsteva.io.postgres.PostgresData]
"""
tables_to_load = {
    "person": [
        "person_id",
        "location_id",
        "year_of_birth",
        "month_of_birth",
        "day_of_birth",
        "birth_datetime",
        "death_datetime",
        "gender_source_value",
        "gender_source_concept_id",
        "status_source_value",
        "cdm_source",
    ],
    "visit_occurrence": [
        "visit_occurrence_id",
        "person_id",
        "visit_occurrence_source_value",
        "preceding_visit_occurrence_id",
        "care_site_id",
        "visit_start_datetime",
        "visit_end_datetime",
        "visit_source_value",
        "visit_source_concept_id",
        "visit_type_source_value",
        "visit_type_source_concept_id",
        "admitted_from_source_value",
        "admitted_from_source_concept_id",
        "discharge_to_source_value",
        "discharge_to_source_concept_id",
        "row_status_source_value",
        "stay_source_value",
        "stay_source_concept_id",
        "cdm_source",
    ],
    "care_site": [
        "care_site_id",
        "care_site_source_value",
        "care_site_name",
        "care_site_short_name",
        "place_of_service_source_value",
        "care_site_type_source_value",
        "valid_start_date",
        "valid_end_date",
    ],
    "visit_detail": [
        "visit_detail_id",
        "visit_occurrence_id",
        "person_id",
        "preceding_visit_detail_id",
        "visit_detail_parent_id",
        "care_site_id",
        "visit_detail_start_date",
        "visit_detail_start_datetime",
        "visit_detail_end_date",
        "visit_detail_end_datetime",
        "visit_detail_source_value",
        "visit_detail_source_concept_id",
        "visit_detail_type_source_value",
        "visit_detail_type_source_concept_id",
        "admitted_from_source_value",
        "admitted_from_source_concept_id",
        "discharge_to_source_value",
        "discharge_to_source_concept_id",
        "cdm_source",
    ],
    "condition_occurrence": [
        "condition_occurrence_id",
        "person_id",
        "visit_occurrence_id",
        "visit_detail_id",
        "condition_start_datetime",
        "condition_source_value",
        "condition_source_concept_id",
        "condition_status_source_value",
        "condition_status_source_concept_id",
        "cdm_source",
    ],
    "procedure_occurrence": [
        "procedure_occurrence_id",
        "person_id",
        "visit_occurrence_id",
        "visit_detail_id",
        "procedure_datetime",
        "procedure_source_value",
        "procedure_source_concept_id",
        "cdm_source",
    ],
    "concept": [
        "concept_id",
        "concept_name",
        "domain_id",
        "vocabulary_id",
        "concept_class_id",
        "standard_concept",
        "concept_code",
        "valid_start_date",
        "valid_end_date",
        "invalid_reason",
    ],
}


# make sure we know how to load the tables we want to save
assert all(table in tables_to_load.keys() for table in default_tables_to_save)

"""
The available tables from i2b2 and their mapping to OMOP standard.
"""
# Tables for each base
i2b2_tables = {
    "edsprod": {
        "visit_occurrence": "i2b2_orbis_visit_dim",
        "note": "i2b2_observation_fact_document",
        "note_ref": "orbis_document",
        "person": "i2b2_patient_dim",
        "condition_occurrence": "i2b2_arem_observation_fact_cim10",
        "procedure_occurrence": "i2b2_arem_observation_fact_ccam",
        "care_site_ref": "orbis_ref_struct_list",
        "care_site": "i2b2_observation_fact_ufr",
        "visit_detail": "i2b2_observation_fact_ufr",
        "biology": "i2b2_observation_fact_lab",
        "fact_relationship": "i2b2_observation_fact_ufr",
        "concept": "orbis_form_ref_concept_list",
    },
    "cse": {
        "visit_occurrence": "i2b2_visit",
        "note": "i2b2_observation_doc",
        "person": "i2b2_patient",
        "condition_occurrence": "i2b2_observation_cim10",
        "procedure_occurrence": "i2b2_observation_ccam",
        "care_site": "i2b2_observation_ufr",
        "visit_detail": "i2b2_observation_ufr",
        "biology": "i2b2_observation_lab",
        "fact_relationship": "i2b2_observation_ufr",
        "concept": "i2b2_concept",
    },
}

# Mapping between i2b2 and OMOP
i2b2_renaming = {
    "note": {
        "instance_num": "note_id",
        "start_date": "note_datetime",
        "concept_cd": "note_class_source_value",
        "encounter_num": "visit_occurrence_id",
        "patient_num": "person_id",
        "observation_blob": "note_text",
        "sourcesystem_cd": "cdm_source",
        "i2b2_action": "row_status_source_value",
    },
    "note_ref": {
        "ids_eds": "note_id",
        "id_dm_doc_ufr": "ufr_source_value",
        "id_dm_doc_us": "us_source_value",
    },
    "person": {
        "patient_num": "person_id",
        "birth_date": "birth_datetime",
        "death_date": "death_datetime",
        "sex_cd": "gender_source_value",
        "sourcesystem_cd": "cdm_source",
    },
    "visit_occurrence": {
        "patient_num": "person_id",
        "encounter_num": "visit_occurrence_id",
        "visit_blob": "visit_occurrence_source_value",
        "start_date": "visit_start_datetime",
        "end_date": "visit_end_datetime",
        "sourcesystem_cd": "cdm_source",
        "type_visite": "visit_source_value",
        "mode_entree": "admitted_from_source_value",
        "mode_sortie": "discharge_to_source_value",
        "i2b2_action": "row_status_source_value",
    },
    "condition_occurrence": {
        "patient_num": "person_id",
        "encounter_num": "visit_occurrence_id",
        "instance_num": "condition_occurrence_id",
        "start_date": "condition_start_datetime",
        "concept_cd": "condition_source_value",
        "tval_char": "condition_status_source_value",
        "location_cd": "care_site_source_value",
        "sourcesystem_cd": "cdm_source",
    },
    "procedure_occurrence": {
        "patient_num": "person_id",
        "encounter_num": "visit_occurrence_id",
        "instance_num": "procedure_occurrence_id",
        "start_date": "procedure_start_datetime",
        "concept_cd": "procedure_source_value",
        "sourcesystem_cd": "cdm_source",
    },
    "care_site": {"location_cd": "care_site_source_value"},
    "care_site_ref": {
        "ids_eds": "care_site_id",
        "id_ref_stuct": "care_site_source_value",
    },
    "visit_detail": {
        "instance_num": "visit_detail_id",
        "encounter_num": "visit_occurrence_id",
        "location_cd": "care_site_id",
        "patient_num": "person_id",
        "start_date": "visit_detail_start_datetime",
        "end_date": "visit_detail_end_datetime",
        "sourcesystem_cd": "cdm_source",
        "i2b2_action": "row_status_source_value",
    },
    "biology": {
        "patient_num": "person_id",
        "encounter_num": "visit_occurrence_id",
        "instance_num": "biology_occurrence_id",
        "start_date": "biology_start_datetime",
        "concept_cd": "biology_source_value",
        "sourcesystem_cd": "cdm_source",
        "nval_num": "biology_status_source_value",
    },
    "fact_relationship": {
        "location_cd": "care_site_source_value",
        "sourcesystem_cd": "cdm_source",
    },
    "concept": {"cd_concept": "concept_id", "lib_concept": "concept_name"},
}

sex_cd_mapping = {"W": "f", "M": "m"}

dict_code_UFR = {
    "014": "APR",
    "028": "ABC",
    "095": "AVC",
    "005": "BJN",
    "009": "BRK",
    "010": "BCT",
    "011": "BCH",
    "033": "BRT",
    "016": "BRC",
    "042": "CFX",
    "019": "CRC",
    "021": "CCH",
    "022": "CCL",
    "029": "ERX",
    "036": "GCL",
    "075": "EGP",
    "038": "HND",
    "026": "HMN",
    "099": "HAD",
    "041": "HTD",
    "032": "JVR",
    "044": "JFR",
    "047": "LRB",
    "049": "LRG",
    "053": "LMR",
    "061": "NCK",
    "096": "PBR",
    "066": "PSL",
    "068": "RPC",
    "069": "RMB",
    "070": "RDB",
    "072": "RTH",
    "073": "SAT",
    "079": "SPR",
    "076": "SLS",
    "084": "SSL",
    "087": "TNN",
    "088": "TRS",
    "090": "VGR",
    "064": "VPD",
    "INC": "INCONNU",
}

visit_type_mapping = {
    "I": "hospitalisés",
    "II": "hospitalisation incomplète",
    "U": "urgence",
    "O": "consultation externe",
}

care_site_level_mapping = {
    30: "Hôpital",
    35: "Unité de consultation (UC)",
    45: "Pôle/DMU",
    50: "Unité Fonctionnelle (UF)",
    70: "Unité d’hébergement (UH)",
    100: "Unité de consultation (UC)",
}



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.filter_df import convert_uf_to_pole
from edsteva.probes.utils.prepare_df import (
    prepare_care_site,
    prepare_note,
    prepare_note_care_site,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_visit(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    extra_data: Data,
    length_of_stays: List[float],
    note_types: Union[str, Dict[str, str]],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_visit`` algorithm computes $c_(t)$ the availability of clinical documents linked to patients' administrative stays:

    $$
    c(t) = \frac{n_{with\,doc}(t)}{n_{visit}(t)}
    $$

    Where $n_{visit}(t)$ is the number of administrative stays, $n_{with\,doc}$ the number of visits having at least one document and $t$ is the month.
    """

    self._metrics = ["c", "n_visit", "n_visit_with_note"]
    check_tables(data=data, required_tables=["note"])

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        start_date=start_date,
        end_date=end_date,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    )

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_relationship=care_site_relationship,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
    )

    note = prepare_note(data, note_types)

    hospital_visit = get_hospital_visit(note, visit_occurrence, care_site)
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    note_predictor_by_level = {hospital_name: hospital_visit}

    # UF selection
    if not hospital_only(care_site_levels=care_site_levels):
        if extra_data:  # pragma: no cover
            visit_detail = prepare_visit_detail(data, start_date, end_date)

            uf_visit, uc_visit, uh_visit = get_visit_detail(
                extra_data=extra_data,
                note=note,
                visit_occurrence=visit_occurrence,
                visit_detail=visit_detail,
                care_site=care_site,
            )
            uf_name = CARE_SITE_LEVEL_NAMES["UF"]
            note_predictor_by_level[uf_name] = uf_visit
            uc_name = CARE_SITE_LEVEL_NAMES["UC"]
            note_predictor_by_level[uc_name] = uc_visit
            uh_name = CARE_SITE_LEVEL_NAMES["UH"]
            note_predictor_by_level[uh_name] = uh_visit

            pole_visit = get_pole_visit(uf_visit, care_site, care_site_relationship)
            pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
            note_predictor_by_level[pole_name] = pole_visit
        else:
            logger.info("Note data are only available at hospital level")
            care_site_levels = ["Hospital"]

    # Concatenate all predictors
    note_predictor = concatenate_predictor_by_level(
        predictor_by_level=note_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    if is_koalas(note_predictor):
        note_predictor.spark.cache()

    return compute_completeness(self, note_predictor)


def compute_completeness(
    self,
    note_predictor: DataFrame,
):
    # Visit with note
    partition_cols = [*self._index.copy(), "date"]
    n_visit_with_note = (
        note_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"has_note": "count"})
        .rename(columns={"has_note": "n_visit_with_note"})
    )
    n_visit_with_note = to("pandas", n_visit_with_note)
    n_visit_with_note = n_visit_with_note[n_visit_with_note.n_visit_with_note > 0]
    n_visit_with_note = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit_with_note,
        partition_cols=partition_cols,
    )

    # Visit total
    note_columns = ["note_type"]
    partition_cols = list(set(partition_cols) - set(note_columns))
    n_visit = (
        note_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )
    n_visit = to("pandas", n_visit)
    n_visit = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit,
        partition_cols=partition_cols,
    )
    note_predictor = n_visit_with_note.merge(
        n_visit,
        on=partition_cols,
    )

    # Compute completeness
    note_predictor["c"] = note_predictor["n_visit"].where(
        note_predictor["n_visit"] == 0,
        note_predictor["n_visit_with_note"] / note_predictor["n_visit"],
    )

    return note_predictor


def get_hospital_visit(
    note: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    note_hospital = note[["visit_occurrence_id", "note_type"]].drop_duplicates()
    note_hospital["has_note"] = True
    hospital_visit = visit_occurrence.merge(
        note_hospital, on="visit_occurrence_id", how="left"
    )
    hospital_visit = hospital_visit.rename(columns={"visit_occurrence_id": "visit_id"})
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")
    if is_koalas(hospital_visit):
        hospital_visit = hospital_visit.spark.cache()

    return hospital_visit


def get_visit_detail(
    extra_data: Data,
    note: DataFrame,
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):  # pragma: no cover
    visit_detail = visit_detail.merge(
        visit_occurrence[
            visit_occurrence.columns.intersection(
                set(["visit_occurrence_id", "length_of_stay", "stay_type"])
            )
        ],
        on="visit_occurrence_id",
    )

    note_detail = prepare_note_care_site(extra_data=extra_data, note=note)
    note_detail = note_detail[
        ["visit_occurrence_id", "note_type", "care_site_id"]
    ].drop_duplicates()
    note_detail["has_note"] = True
    note_detail = visit_detail.merge(
        note_detail,
        on=["visit_occurrence_id", "care_site_id"],
        how="left",
    ).drop(columns="visit_occurrence_id")

    note_detail = note_detail.merge(care_site, on="care_site_id")

    uf_name = CARE_SITE_LEVEL_NAMES["UF"]
    uf_visit = note_detail[note_detail["care_site_level"] == uf_name]
    uc_name = CARE_SITE_LEVEL_NAMES["UC"]
    uc_visit = note_detail[note_detail["care_site_level"] == uc_name]
    uh_name = CARE_SITE_LEVEL_NAMES["UH"]
    uh_visit = note_detail[note_detail["care_site_level"] == uh_name]

    if is_koalas(note_detail):
        uf_visit = uf_visit.spark.cache()
        uc_visit = uc_visit.spark.cache()
        uh_visit = uh_visit.spark.cache()

    return uf_visit, uc_visit, uh_visit


def get_pole_visit(
    uf_visit: DataFrame,
    care_site: DataFrame,
    care_site_relationship: DataFrame,
):  # pragma: no cover
    care_site_cols = list(
        set(
            [
                "care_site_short_name",
                "care_site_level",
                "care_site_specialty",
                "specialties_set",
                "care_sites_set",
            ]
        ).intersection(uf_visit.columns)
    )
    pole_visit = convert_uf_to_pole(
        table=uf_visit.drop(columns=care_site_cols),
        table_name="uf_visit",
        care_site_relationship=care_site_relationship,
    )

    pole_visit = pole_visit.merge(care_site, on="care_site_id")
    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    pole_visit = pole_visit[pole_visit["care_site_level"] == pole_name]
    if is_koalas(pole_visit):
        pole_visit = pole_visit.spark.cache()

    return pole_visit



from .estimates_densities_plot import get_estimates_densities_plot_config
from .normalized_probe_dashboard import get_normalized_probe_dashboard_config
from .normalized_probe_plot import get_normalized_probe_plot_config
from .probe_dashboard import get_probe_dashboard_config
from .probe_plot import get_probe_plot_config



from edsteva.utils.typing import DataFrame

from .defaults import (
    get_c_0_min_selection,
    get_error_max_selection,
    get_t_0_selection,
    horizontal_max_error,
    horizontal_min_c0,
    normalized_model_line,
    normalized_probe_line,
)


def get_normalized_probe_dashboard_config(self, predictor: DataFrame):
    c_0_min_selection, c_0_min_filter = get_c_0_min_selection(predictor=predictor)
    t_0_selection, t_0_min_filter = get_t_0_selection(predictor=predictor)
    error_max_selection, error_max_filter = get_error_max_selection(predictor=predictor)
    return dict(
        estimates_selections=[c_0_min_selection, t_0_selection, error_max_selection],
        estimates_filters=[c_0_min_filter, t_0_min_filter, error_max_filter],
        probe_line=normalized_probe_line,
        model_line=normalized_model_line,
        extra_horizontal_bar_charts=[horizontal_min_c0, horizontal_max_error],
        extra_vertical_bar_charts=[],
    )



import numpy as np
import pandas as pd
from loguru import logger


def generate_note(
    generator: np.random.Generator,
    visit_care_site: pd.DataFrame,
    note_type: str,
    note_date_col: str,
    id_visit_col: str,
    note_type_col: str,
    t0_visit: int,
    care_site_id: int,
    date_col: str,
    mode: str,
):
    if mode == "step":
        return _generate_note_step(
            generator=generator,
            visit_care_site=visit_care_site,
            note_type=note_type,
            note_date_col=note_date_col,
            id_visit_col=id_visit_col,
            note_type_col=note_type_col,
            t0_visit=t0_visit,
            care_site_id=care_site_id,
            date_col=date_col,
        )

    if mode == "rect":
        return _generate_note_rect(
            generator=generator,
            visit_care_site=visit_care_site,
            note_type=note_type,
            note_date_col=note_date_col,
            id_visit_col=id_visit_col,
            note_type_col=note_type_col,
            t0_visit=t0_visit,
            care_site_id=care_site_id,
            date_col=date_col,
        )


def _generate_note_step(
    generator: np.random.Generator,
    visit_care_site,
    note_type,
    care_site_id,
    date_col,
    note_date_col,
    id_visit_col,
    note_type_col,
    t0_visit,
):
    t_end = visit_care_site[date_col].max()
    t0 = generator.integers(t0_visit, t_end)
    c_before = generator.uniform(0, 0.2)
    c_after = generator.uniform(0.8, 1)
    note_before_t0_visit = (
        visit_care_site[visit_care_site[date_col] <= t0_visit][[id_visit_col, date_col]]
        .sample(frac=c_before)
        .rename(columns={date_col: note_date_col})
    )
    # Stratify visit between t0_visit and t0 to
    # ensure that these elements are represented
    # in the final notes dataset.
    note_before_t0 = (
        visit_care_site[
            (visit_care_site[date_col] <= t0) & (visit_care_site[date_col] > t0_visit)
        ][[id_visit_col, date_col]]
        .sample(frac=c_before)
        .rename(columns={date_col: note_date_col})
    )

    note_after_t0 = (
        visit_care_site[visit_care_site[date_col] > t0][[id_visit_col, date_col]]
        .sample(frac=c_after)
        .rename(columns={date_col: note_date_col})
    )

    note = pd.concat([note_before_t0_visit, note_before_t0, note_after_t0])

    note[note_date_col] = pd.to_datetime(note[note_date_col], unit="s")
    note[note_type_col] = note_type
    note["care_site_id"] = care_site_id
    note["t_0"] = t0
    logger.debug("Generate synthetic note deploying as step function")

    return note


def _generate_note_rect(
    generator: np.random.Generator,
    visit_care_site,
    note_type,
    care_site_id,
    date_col,
    note_date_col,
    id_visit_col,
    note_type_col,
    t0_visit,
):
    t1_visit = visit_care_site["t_1_min"].max()
    t0 = generator.integers(t0_visit, t0_visit + (t1_visit - t0_visit) / 3)
    t1 = generator.integers(t0_visit + 2 * (t1_visit - t0_visit) / 3, t1_visit)
    c_out = generator.uniform(0, 0.1)
    c_in = generator.uniform(0.8, 1)

    note_before_t0 = (
        visit_care_site[visit_care_site[date_col] <= t0][[id_visit_col, date_col]]
        .sample(frac=c_out)
        .rename(columns={date_col: note_date_col})
    )
    note_between_t0_t1 = (
        visit_care_site[
            (visit_care_site[date_col] > t0) & (visit_care_site[date_col] <= t1)
        ][[id_visit_col, date_col]]
        .sample(frac=c_in)
        .rename(columns={date_col: note_date_col})
    )

    note_after_t1 = (
        visit_care_site[(visit_care_site[date_col] > t1)][[id_visit_col, date_col]]
        .sample(frac=c_out)
        .rename(columns={date_col: note_date_col})
    )

    note = pd.concat(
        [
            note_before_t0,
            note_between_t0_t1,
            note_after_t1,
        ]
    )

    note[note_date_col] = pd.to_datetime(note[note_date_col], unit="s")
    note[note_type_col] = note_type
    note["care_site_id"] = care_site_id
    note["t_0"] = t0
    note["t_1"] = t1
    logger.debug("Generate synthetic note deploying as rectangle function")

    return note



from typing import List

import numpy as np
import pandas as pd

from edsteva.utils.checks import check_columns


def quantile(
    predictor: pd.DataFrame,
    index: List[str],
    q: float = 0.8,
    x: str = "date",
    y: str = "c",
    threshold: str = "c_0",
):
    estimates = c_0_from_quantile(predictor=predictor, index=index, q=q, x=x, y=y)
    return t_0_from_c_0(predictor=estimates, index=index, x=x, y=y, threshold=threshold)


def c_0_from_quantile(
    predictor: pd.DataFrame,
    index: List[str],
    q: float = 0.8,
    x: str = "date",
    y: str = "c",
) -> pd.DataFrame:
    r"""Compute the quantile on the given y-axis. Column $c_0$ is created.

    $$
    \hat{c_0} = x^{th} \text{ quantile of } c(t)
    $$

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    index : List[str]
        Variable from which data is grouped

        **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
    q : float, optional
        Quantile value
    x : str, optional
        Column name for the time variable $t$
    y : str, optional
        Column name  for the completeness variable $c(t)$
    """

    check_columns(df=predictor, required_columns=[*index, x, y])

    quantile = (
        predictor.groupby(index)[[y]]
        .agg(lambda g: np.quantile(g, q=q))
        .rename(columns={y: "c_0"})
    )

    return predictor.merge(quantile, on=index)


def t_0_from_c_0(
    predictor: pd.DataFrame,
    index: List[str],
    x: str = "date",
    y: str = "c",
    threshold: str = "c_0",
) -> pd.DataFrame:
    r"""Compute $t_0$ column using value of $c_0$

    Returns the first date at which values are greater than $c_0$:

    $$
    \hat{t_0} = \underset{t}{\mathrm{argmin}}(c(t) \geq \hat{c_0})
    $$

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    index : List[str]
        Variable from which data is grouped
    x : str, optional
        Column name for the time variable $t$
    y : str, optional
        Column name  for the completeness variable $c(t)$
    threshold : str, optional
        Column name  for the threshold variable $t_0$
    """

    check_columns(df=predictor, required_columns=[*index, x, y, threshold])

    threshold = (
        predictor[predictor[y] > predictor[threshold]]
        .groupby(index)[[x]]
        .min()
        .rename(columns={x: "t_0"})
    )

    return predictor.merge(threshold, on=index)



import catalogue

from edsteva.probes.note.viz_configs import n_note, per_note, per_visit

normalized_probe_dashboard = catalogue.create(
    "edsteva.probes.note.viz_configs", "normalized_probe_dashboard"
)
normalized_probe_dashboard.register(
    "per_visit_default", func=per_visit.get_normalized_probe_dashboard_config
)
normalized_probe_dashboard.register(
    "per_note_default", func=per_note.get_normalized_probe_dashboard_config
)
normalized_probe_dashboard.register(
    "n_note", func=n_note.get_normalized_probe_dashboard_config
)
probe_dashboard = catalogue.create("edsteva.probes.note.viz_configs", "probe_dashboard")
probe_dashboard.register("per_visit_default", func=per_visit.get_probe_dashboard_config)
probe_dashboard.register("per_note_default", func=per_note.get_probe_dashboard_config)
probe_dashboard.register("n_note", func=n_note.get_probe_dashboard_config)

estimates_densities_plot = catalogue.create(
    "edsteva.probes.note.viz_configs", "estimates_densities_plot"
)
estimates_densities_plot.register(
    "per_visit_default", func=per_visit.get_estimates_densities_plot_config
)
estimates_densities_plot.register(
    "per_note_default", func=per_note.get_estimates_densities_plot_config
)
estimates_densities_plot.register(
    "n_note", func=n_note.get_estimates_densities_plot_config
)

normalized_probe_plot = catalogue.create(
    "edsteva.probes.note.viz_configs", "normalized_probe_plot"
)
normalized_probe_plot.register(
    "per_visit_default", func=per_visit.get_normalized_probe_plot_config
)
normalized_probe_plot.register(
    "per_note_default", func=per_note.get_normalized_probe_plot_config
)
normalized_probe_plot.register("n_note", func=n_note.get_normalized_probe_plot_config)

probe_plot = catalogue.create("edsteva.probes.note.viz_configs", "probe_plot")
probe_plot.register("per_visit_default", func=per_visit.get_probe_plot_config)
probe_plot.register("per_note_default", func=per_note.get_probe_plot_config)
probe_plot.register("n_note", func=n_note.get_probe_plot_config)

viz_configs = dict(
    normalized_probe_dashboard=normalized_probe_dashboard,
    probe_dashboard=probe_dashboard,
    estimates_densities_plot=estimates_densities_plot,
    normalized_probe_plot=normalized_probe_plot,
    probe_plot=probe_plot,
)



from datetime import datetime
from typing import Dict, List, Tuple, Union

import pandas as pd
from loguru import logger

from edsteva.utils.checks import check_columns, check_tables
from edsteva.utils.framework import get_framework, is_koalas, to
from edsteva.utils.typing import Data, DataFrame

from .filter_df import (
    filter_table_by_care_site,
    filter_table_by_date,
    filter_table_by_length_of_stay,
    filter_table_by_type,
    filter_valid_observations,
)


def prepare_visit_occurrence(
    data: Data,
    stay_types: Union[str, Dict[str, str]],
    length_of_stays: List[float],
    start_date: datetime = None,
    end_date: datetime = None,
):
    required_columns = [
        "visit_occurrence_id",
        "visit_source_value",
        "visit_start_datetime",
        "visit_end_datetime",
        "care_site_id",
        "row_status_source_value",
        "visit_occurrence_source_value",
    ]
    check_columns(
        data.visit_occurrence,
        required_columns=required_columns,
        df_name="visit_occurrence",
    )
    visit_occurrence = data.visit_occurrence[required_columns]

    visit_occurrence = filter_valid_observations(
        table=visit_occurrence,
        table_name="visit_occurrence",
        invalid_naming="supprimé",
    )

    if length_of_stays:
        visit_occurrence = filter_table_by_length_of_stay(
            visit_occurrence=visit_occurrence, length_of_stays=length_of_stays
        )

    visit_occurrence = visit_occurrence.rename(
        columns={"visit_source_value": "stay_type", "visit_start_datetime": "date"}
    )

    visit_occurrence = filter_table_by_date(
        table=visit_occurrence,
        table_name="visit_occurrence",
        start_date=start_date,
        end_date=end_date,
    )

    if stay_types:
        visit_occurrence = filter_table_by_type(
            table=visit_occurrence,
            table_name="visit_occurrence",
            type_groups=stay_types,
            source_col="stay_type",
            target_col="stay_type",
        )

    return visit_occurrence


def prepare_measurement(
    data: Data,
    biology_relationship: pd.DataFrame,
    concept_codes: List[str],
    concepts_sets: Union[str, Dict[str, str]],
    root_terminology: str,
    standard_terminologies: List[str],
    per_visit: bool,
    start_date: datetime = None,
    end_date: datetime = None,
):
    measurement_columns = [
        "measurement_id",
        "visit_occurrence_id",
        "measurement_datetime",
        "row_status_source_value",
        "measurement_source_concept_id",
    ]

    check_columns(
        data.measurement,
        required_columns=measurement_columns,
        df_name="measurement",
    )
    measurement = data.measurement[measurement_columns].rename(
        columns={
            "measurement_source_concept_id": "{}_concept_id".format(root_terminology),
            "measurement_datetime": "date",
        }
    )
    measurement = filter_valid_observations(
        table=measurement, table_name="measurement", valid_naming="Validé"
    )

    biology_relationship = biology_relationship[
        [
            "{}_{}".format(root_terminology, concept_col)
            for concept_col in ["concept_id", "concept_code", "concept_name"]
        ]
        + [
            "{}_{}".format(terminology, concept_col)
            for terminology in standard_terminologies
            for concept_col in ["concept_code", "concept_name", "vocabulary"]
        ]
    ]
    biology_relationship = to(get_framework(measurement), biology_relationship)
    if is_koalas(biology_relationship):
        biology_relationship = biology_relationship.spark.hint("broadcast")
    measurement = measurement.merge(
        biology_relationship, on="{}_concept_id".format(root_terminology)
    )
    if not per_visit:
        measurement = filter_table_by_date(
            table=measurement,
            table_name="measurement",
            start_date=start_date,
            end_date=end_date,
        )

    if concept_codes:
        measurement_by_terminology = []
        for standard_terminology in standard_terminologies:
            measurement_by_terminology.append(
                measurement[
                    measurement["{}_concept_code".format(standard_terminology)].isin(
                        concept_codes
                    )
                ]
            )
        measurement = get_framework(measurement).concat(measurement_by_terminology)

    if concepts_sets:
        measurement_by_terminology = []
        for standard_terminology in standard_terminologies:
            measurement_by_terminology.append(
                filter_table_by_type(
                    table=measurement,
                    table_name="measurement",
                    type_groups=concepts_sets,
                    source_col="{}_concept_code".format(standard_terminology),
                    target_col="concepts_set",
                )
            )
        measurement = get_framework(measurement).concat(measurement_by_terminology)
    return measurement


def prepare_condition_occurrence(
    data: Data,
    extra_data: Data,
    visit_occurrence: DataFrame,
    source_systems: List[str],
    diag_types: Union[str, Dict[str, str]],
    condition_types: Union[str, Dict[str, str]],
    start_date: datetime = None,
    end_date: datetime = None,
):
    condition_occurrence_tables = []
    if "AREM" in source_systems:  # pragma: no cover
        check_tables(
            data=extra_data,
            required_tables=["visit_occurrence", "condition_occurrence"],
        )
        # Fetch conditions from Data lake
        I2B2_visit = extra_data.visit_occurrence[
            ["visit_occurrence_id", "visit_occurrence_source_value"]
        ]
        I2B2_condition_occurrence = extra_data.condition_occurrence[
            [
                "visit_occurrence_id",
                "condition_occurrence_id",
                "condition_status_source_value",
                "condition_start_datetime",
                "condition_source_value",
                "care_site_source_value",
                "cdm_source",
            ]
        ]
        # Add visit_occurrence_source_value
        arem_condition_occurrence = I2B2_visit.merge(
            I2B2_condition_occurrence,
            on="visit_occurrence_id",
            how="inner",
        ).drop(columns="visit_occurrence_id")

        # Link with visit_occurrence_source_value
        arem_condition_occurrence = arem_condition_occurrence.merge(
            visit_occurrence[["visit_occurrence_source_value", "visit_occurrence_id"]],
            on="visit_occurrence_source_value",
        ).drop(columns="visit_occurrence_source_value")
        arem_condition_occurrence = arem_condition_occurrence[
            arem_condition_occurrence.cdm_source == "AREM"
        ]
        arem_condition_occurrence["visit_detail_id"] = None
        condition_occurrence_tables.append(arem_condition_occurrence)

    if "ORBIS" in source_systems:
        orbis_condition_occurrence = data.condition_occurrence[
            [
                "visit_occurrence_id",
                "condition_occurrence_id",
                "visit_detail_id",
                "condition_source_value",
                "condition_start_datetime",
                "condition_status_source_value",
                "row_status_source_value",
                "cdm_source",
            ]
        ]
        orbis_condition_occurrence = filter_valid_observations(
            table=orbis_condition_occurrence,
            table_name="orbis_condition_occurrence",
            valid_naming="Actif",
        )
        orbis_condition_occurrence = orbis_condition_occurrence[
            orbis_condition_occurrence.cdm_source == "ORBIS"
        ]
        condition_occurrence_tables.append(orbis_condition_occurrence)

    framework = get_framework(condition_occurrence_tables[0])
    condition_occurrence = framework.concat(
        condition_occurrence_tables, ignore_index=True
    )

    # Filter date
    condition_occurrence = condition_occurrence.rename(
        columns={"condition_start_datetime": "date"}
    )
    condition_occurrence = filter_table_by_date(
        table=condition_occurrence,
        table_name="condition_occurrence",
        start_date=start_date,
        end_date=end_date,
    )

    # Filter source system
    condition_occurrence = condition_occurrence.rename(
        columns={"cdm_source": "source_system"}
    )

    # Filter diagnostics
    condition_occurrence = condition_occurrence.rename(
        columns={"condition_status_source_value": "diag_type"}
    )
    if diag_types:
        condition_occurrence = filter_table_by_type(
            table=condition_occurrence,
            table_name="condition_occurrence",
            type_groups=diag_types,
            source_col="diag_type",
            target_col="diag_type",
        )

    # Filter conditions
    condition_occurrence = condition_occurrence.rename(
        columns={"condition_source_value": "condition_type"}
    )
    if condition_types:
        condition_occurrence = filter_table_by_type(
            table=condition_occurrence,
            table_name="condition_occurrence",
            type_groups=condition_types,
            source_col="condition_type",
            target_col="condition_type",
        )

    return condition_occurrence


def prepare_care_site(
    data: Data,
    care_site_ids: Union[int, List[int]],
    care_site_short_names: Union[str, List[str]],
    care_site_specialties: Union[str, List[str]],
    specialties_sets: Union[str, Dict[str, str]],
    care_sites_sets: Union[str, Dict[str, str]],
    care_site_relationship: pd.DataFrame,
):
    care_site = data.care_site[
        [
            "care_site_id",
            "care_site_type_source_value",
            "care_site_short_name",
            "place_of_service_source_value",
        ]
    ]
    care_site = care_site.rename(
        columns={
            "care_site_type_source_value": "care_site_level",
            "place_of_service_source_value": "care_site_specialty",
        }
    )
    if care_site_ids or care_site_short_names or care_site_specialties:
        care_site = filter_table_by_care_site(
            table_to_filter=care_site,
            care_site_relationship=care_site_relationship,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
            care_site_specialties=care_site_specialties,
        )

    # Add care_sites_set
    if care_sites_sets:
        care_site = filter_table_by_type(
            table=care_site,
            table_name="care_site",
            type_groups=care_sites_sets,
            source_col="care_site_short_name",
            target_col="care_sites_set",
        )

    # Add specialties_set
    if specialties_sets:
        care_site = filter_table_by_type(
            table=care_site,
            table_name="care_site",
            type_groups=specialties_sets,
            source_col="care_site_specialty",
            target_col="specialties_set",
        )
    return care_site


def prepare_note(
    data: Data,
    note_types: Union[str, Dict[str, str]],
    start_date: datetime = None,
    end_date: datetime = None,
):
    note = data.note[
        [
            "note_id",
            "visit_occurrence_id",
            "note_datetime",
            "note_class_source_value",
            "row_status_source_value",
            "note_text",
        ]
    ]
    note = note.rename(
        columns={"note_class_source_value": "note_type", "note_datetime": "date"}
    )
    note = note[~(note["note_text"].isna())]
    note = note.drop(columns=["note_text"])

    note = filter_table_by_date(
        table=note,
        table_name="note",
        start_date=start_date,
        end_date=end_date,
    )

    note = filter_valid_observations(
        table=note, table_name="note", valid_naming="Actif"
    )

    # Add note type
    if note_types:
        note = filter_table_by_type(
            table=note,
            table_name="note",
            type_groups=note_types,
            source_col="note_type",
            target_col="note_type",
        )

    return note


def prepare_note_care_site(extra_data: Data, note: DataFrame):  # pragma: no cover
    check_tables(
        data=extra_data,
        required_tables=["note_ref", "care_site_ref"],
    )

    note_ref = extra_data.note_ref[
        [
            "note_id",
            "ufr_source_value",
            "us_source_value",
        ]
    ]
    care_site_ref = extra_data.care_site_ref[
        [
            "care_site_source_value",
            "care_site_id",
        ]
    ]

    note_ref = note_ref.melt(
        id_vars="note_id",
        value_name="care_site_source_value",
    )
    note_ref = note_ref.merge(care_site_ref, on="care_site_source_value")
    return note.merge(note_ref[["note_id", "care_site_id"]], on="note_id")


def prepare_visit_detail(
    data: Data,
    start_date: datetime,
    end_date: datetime,
):
    visit_detail = data.visit_detail[
        [
            "visit_detail_id",
            "visit_occurrence_id",
            "visit_detail_start_datetime",
            "visit_detail_type_source_value",
            "care_site_id",
            "row_status_source_value",
        ]
    ]
    visit_detail = visit_detail.rename(
        columns={
            "visit_detail_id": "visit_id",
            "visit_detail_start_datetime": "date",
            "visit_detail_type_source_value": "visit_detail_type",
        }
    )
    visit_detail = filter_valid_observations(
        table=visit_detail, table_name="visit_detail", valid_naming="Actif"
    )

    return filter_table_by_date(
        table=visit_detail,
        table_name="visit_detail",
        start_date=start_date,
        end_date=end_date,
    )


def prepare_care_site_relationship(data: Data) -> pd.DataFrame:
    """Computes hierarchical care site structure

    Parameters
    ----------
    data : Data
        Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]

    Example
    -------

    | care_site_id | care_site_level            | care_site_short_name | parent_care_site_id | parent_care_site_level     | parent_care_site_short_name |
    | :----------- | :------------------------- | :------------------- | :------------------ | :------------------------- | :-------------------------- |
    | 8312056386   | Unité Fonctionnelle (UF)   | UF A                 | 8312027648          | Pôle/DMU                   | Pole A                      |
    | 8312022130   | Pôle/DMU                   | Pole B               | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312016782   | Service/Département        | Service A            | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312010155   | Unité Fonctionnelle (UF)   | UF B                 | 8312022130          | Pôle/DMU                   | Pole B                      |
    | 8312067829   | Unité de consultation (UC) | UC A                 | 8312051097          | Unité de consultation (UC) | UC B                        |

    """
    fact_relationship = data.fact_relationship[
        [
            "fact_id_1",
            "fact_id_2",
            "domain_concept_id_1",
            "relationship_concept_id",
        ]
    ]
    fact_relationship = to("pandas", fact_relationship)

    care_site_relationship = fact_relationship[
        (fact_relationship["domain_concept_id_1"] == 57)  # Care_site domain
        & (fact_relationship["relationship_concept_id"] == 46233688)  # Included in
    ]
    care_site_relationship = care_site_relationship.drop(
        columns=["domain_concept_id_1", "relationship_concept_id"]
    )
    care_site_relationship = care_site_relationship.rename(
        columns={"fact_id_1": "care_site_id", "fact_id_2": "parent_care_site_id"}
    )

    care_site = data.care_site[
        [
            "care_site_id",
            "care_site_type_source_value",
            "care_site_short_name",
            "place_of_service_source_value",
        ]
    ]
    care_site = to("pandas", care_site)
    care_site = care_site.rename(
        columns={
            "care_site_type_source_value": "care_site_level",
            "place_of_service_source_value": "care_site_specialty",
        }
    )
    care_site_relationship = care_site.merge(
        care_site_relationship, on="care_site_id", how="left"
    )

    parent_care_site = care_site.rename(
        columns={
            "care_site_level": "parent_care_site_level",
            "care_site_id": "parent_care_site_id",
            "care_site_short_name": "parent_care_site_short_name",
            "care_site_specialty": "parent_care_site_specialty",
        }
    )
    logger.debug("Create care site relationship to link UC to UF and UF to Pole")

    return care_site_relationship.merge(
        parent_care_site, on="parent_care_site_id", how="left"
    )


def prepare_biology_relationship(
    data: Data,
    standard_terminologies: List[str],
    source_terminologies: Dict[str, str],
    mapping: List[Tuple[str, str, str]],
) -> pd.DataFrame:
    """Computes biology relationship

    Parameters
    ----------
    data : Data
        Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]

    Example
    -------

    | care_site_id | care_site_level            | care_site_short_name | parent_care_site_id | parent_care_site_level     | parent_care_site_short_name |
    | :----------- | :------------------------- | :------------------- | :------------------ | :------------------------- | :-------------------------- |
    | 8312056386   | Unité Fonctionnelle (UF)   | UF A                 | 8312027648          | Pôle/DMU                   | Pole A                      |
    | 8312022130   | Pôle/DMU                   | Pole B               | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312016782   | Service/Département        | Service A            | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312010155   | Unité Fonctionnelle (UF)   | UF B                 | 8312022130          | Pôle/DMU                   | Pole B                      |
    | 8312067829   | Unité de consultation (UC) | UC A                 | 8312051097          | Unité de consultation (UC) | UC B                        |

    """

    logger.debug(
        "Create biology relationship to link ANALYSES LABORATOIRE to ANABIO to LOINC"
    )

    check_tables(data=data, required_tables=["concept", "concept_relationship"])
    concept_columns = [
        "concept_id",
        "concept_name",
        "concept_code",
        "vocabulary_id",
    ]

    concept_relationship_columns = [
        "concept_id_1",
        "concept_id_2",
        "relationship_id",
    ]
    check_columns(
        data.concept,
        required_columns=concept_columns,
        df_name="concept",
    )

    check_columns(
        data.concept_relationship,
        required_columns=concept_relationship_columns,
        df_name="concept_relationship",
    )
    concept = to("pandas", data.concept[concept_columns])
    concept_relationship = to(
        "pandas", data.concept_relationship[concept_relationship_columns]
    )
    concept_by_terminology = {}
    for terminology, regex in source_terminologies.items():
        concept_by_terminology[terminology] = (
            concept[concept.vocabulary_id.str.contains(regex)]
            .rename(
                columns={
                    "concept_id": "{}_concept_id".format(terminology),
                    "concept_name": "{}_concept_name".format(terminology),
                    "concept_code": "{}_concept_code".format(terminology),
                }
            )
            .drop(columns="vocabulary_id")
        )
    root_terminology = mapping[0][0]
    biology_relationship = concept_by_terminology[root_terminology]
    for source, target, relationship_id in mapping:
        relationship = concept_relationship.rename(
            columns={
                "concept_id_1": "{}_concept_id".format(source),
                "concept_id_2": "{}_concept_id".format(target),
            }
        )[concept_relationship.relationship_id == relationship_id].drop(
            columns="relationship_id"
        )
        relationship = relationship.merge(
            concept_by_terminology[target], on="{}_concept_id".format(target)
        )
        biology_relationship = biology_relationship.merge(
            relationship, on="{}_concept_id".format(source), how="left"
        )

    # Get ITM code in priority and if not get GLIMS code
    for standard_terminology in standard_terminologies:
        biology_relationship[
            "{}_concept_code".format(standard_terminology)
        ] = biology_relationship[
            "{}_ITM_concept_code".format(standard_terminology)
        ].mask(
            biology_relationship[
                "{}_ITM_concept_code".format(standard_terminology)
            ].isna(),
            biology_relationship["GLIMS_{}_concept_code".format(standard_terminology)],
        )
        biology_relationship[
            "{}_concept_name".format(standard_terminology)
        ] = biology_relationship[
            "{}_ITM_concept_name".format(standard_terminology)
        ].mask(
            biology_relationship[
                "{}_ITM_concept_name".format(standard_terminology)
            ].isna(),
            biology_relationship["GLIMS_{}_concept_name".format(standard_terminology)],
        )
        biology_relationship["{}_vocabulary".format(standard_terminology)] = "ITM"
        biology_relationship[
            "{}_vocabulary".format(standard_terminology)
        ] = biology_relationship["{}_vocabulary".format(standard_terminology)].mask(
            biology_relationship[
                "{}_ITM_concept_code".format(standard_terminology)
            ].isna(),
            "GLIMS",
        )
    return biology_relationship



from typing import Union

import pandas
from databricks import koalas

from edsteva.io import HiveData, LocalData, PostgresData, SyntheticData

DataFrame = Union[koalas.DataFrame, pandas.DataFrame]
Series = Union[koalas.Series, pandas.Series]
DataObject = Union[DataFrame, Series]
Data = Union[HiveData, PostgresData, LocalData, SyntheticData]



from typing import Callable, List

import numpy as np
import pandas as pd

from edsteva.utils.checks import check_columns
from edsteva.utils.loss_functions import l2_loss


def loss_minimization(
    predictor: pd.DataFrame,
    index: List[str],
    x_col: str = "date",
    y_col: str = "c",
    loss_function: Callable = l2_loss,
) -> pd.DataFrame:
    r"""Computes the threshold $t_0$ of a predictor $c(t)$ by minimizing the following loss function:

    $$
    \begin{aligned}
    \mathcal{L}(t_0) & = \frac{\sum_{t = t_{min}}^{t_{max}} \mathcal{l}(c(t), f_{t_0}(t))}{t_{max} - t_{min}} \\
    \hat{t_0} & = \underset{t_0}{\mathrm{argmin}}(\mathcal{L}(t_0))
    \end{aligned}
    $$

    Where the loss function $\mathcal{l}$ is by default the L2 distance and the estimated completeness $c_0$ is the mean completeness after $t_0$.

    $$
    \begin{aligned}
    \mathcal{l}(c(t), f_{t_0}(t)) & = |c(t) - f_{t_0}(t)|^2 \\
    c_0 & = \frac{\sum_{t = t_0}^{t_{max}} c(t)}{t_{max} - t_0}
    \end{aligned}
    $$


    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    index : List[str]
        Variable from which data is grouped

        **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
    x_col : str, optional
        Column name for the time variable $t$
    y_col : str, optional
        Column name  for the completeness variable $c(t)$
    loss_function : Callable, optional
        The loss function $\mathcal{L}$
    """
    check_columns(df=predictor, required_columns=[*index, x_col, y_col])
    predictor = predictor.sort_values(x_col)
    cols = [*index, x_col, y_col]
    iter = predictor[cols].groupby(index)
    results = []
    for partition, group in iter:
        row = dict(zip(index, partition))
        t_0, c_0 = _compute_one_threshold(
            group,
            x_col,
            y_col,
            loss_function,
        )
        row["t_0"] = t_0
        row["c_0"] = c_0
        results.append(row)

    return pd.DataFrame(results)


def _compute_one_threshold(
    group: pd.DataFrame,
    x_col: str,
    y_col: str,
    loss_func: Callable,
):
    target = group[[x_col, y_col]].to_numpy()
    best_loss, best_x0, best_y0 = np.inf, None, None
    for idx in range(len(target)):
        x0 = target[idx, 0]
        y_before_x0 = target[:idx, 1]
        y_after_x0 = target[idx:, 1]
        y0 = y_after_x0.mean()
        residual = np.hstack([y_before_x0, (y_after_x0 - y0)])
        loss = loss_func(residual).mean()
        if loss < best_loss:
            best_x0 = x0
            best_y0 = y0
            best_loss = loss

    return best_x0, best_y0



import uuid
from copy import deepcopy
from typing import Dict

import altair as alt
import pandas as pd
from IPython.display import HTML, display

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.probes.utils.utils import CARE_SITE_LEVEL_NAMES
from edsteva.viz.utils import (
    add_estimates_filters,
    add_interactive_selection,
    concatenate_charts,
    configure_style,
    create_groupby_selection,
    filter_predictor,
    generate_error_line,
    generate_horizontal_bar_charts,
    generate_main_chart,
    generate_model_line,
    generate_probe_line,
    generate_time_line,
    generate_vertical_bar_charts,
    month_diff,
    save_html,
)


def normalized_probe_dashboard(
    probe: BaseProbe,
    fitted_model: BaseModel,
    care_site_level: str = CARE_SITE_LEVEL_NAMES["Hospital"],
    save_path: str = None,
    x_axis_title: str = None,
    y_axis_title: str = None,
    main_chart_config: Dict[str, str] = None,
    model_line_config: Dict[str, str] = None,
    error_line_config: Dict[str, str] = None,
    probe_line_config: Dict[str, str] = None,
    estimates_selections: Dict[str, str] = None,
    estimates_filters: Dict[str, str] = None,
    vertical_bar_charts_config: Dict[str, str] = None,
    horizontal_bar_charts_config: Dict[str, str] = None,
    time_line_config: Dict[str, str] = None,
    chart_style: Dict[str, float] = None,
    **kwargs,
):
    r"""Displays an interactive chart with:

    - On the top, the aggregated normalized completeness predictor $\frac{c(\Delta t)}{c_0}$ over normalized time $\Delta t = t - t_0$. It represents the overall deviation from the Model.
    - On the bottom, interactive filters including all the columns in the [Probe][probe] (such as time, care site, number of visits...etc.) and all the estimates (coefficients and metrics) in the [Model][model].

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$
    fitted_model : BaseModel
        Model fitted to the probe
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    error_line_config: Dict[str, str], optional
        If not None, configuration used to construct the error line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    estimates_selections: Dict[str, str], optional
        If not None, configuration used to construct the estimates selections.
    estimates_filters: Dict[str, str], optional
        If not None, configuration used to construct the estimates filters.
    vertical_bar_charts_config: Dict[str, str], optional
        If not None, configuration used to construct the vertical bar charts.
    horizontal_bar_charts_config: Dict[str, str], optional
        If not None, configuration used to construct the horizontal bar charts.
    time_line_config: Dict[str, str], optional
        If not None, configuration used to construct the time line.
    chart_style: Dict[str, float], optional
        If not None, configuration used to configure the chart style.
        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    """
    alt.data_transformers.disable_max_rows()

    predictor = probe.predictor.copy()
    estimates = fitted_model.estimates.copy()
    predictor = predictor.merge(estimates, on=probe._index)

    predictor["normalized_date"] = month_diff(
        predictor["date"], predictor["t_0"]
    ).astype(int)
    predictor["normalized_c_0"] = predictor["c_0"].mask(
        (predictor["normalized_date"] < 0) | (predictor["c_0"] == 0), 1
    )

    predictor["model"] = 1
    predictor["model"] = predictor["model"].where(predictor["normalized_date"] >= 0, 0)

    predictor.t_0 = predictor.t_0.dt.strftime("%Y-%m")
    probe_config = deepcopy(probe.get_viz_config("normalized_probe_dashboard"))
    model_config = deepcopy(
        fitted_model.get_viz_config("normalized_probe_dashboard", predictor=predictor)
    )
    if not main_chart_config:
        main_chart_config = probe_config["main_chart"]
    if not time_line_config:
        time_line_config = probe_config["time_line"]
    if not error_line_config:
        error_line_config = probe_config["error_line"]
    if not vertical_bar_charts_config:
        vertical_bar_charts_config = probe_config["vertical_bar_charts"]
        vertical_bar_charts_config["y"] = (
            vertical_bar_charts_config["y"] + model_config["extra_vertical_bar_charts"]
        )
    if not horizontal_bar_charts_config:
        horizontal_bar_charts_config = probe_config["horizontal_bar_charts"]
        horizontal_bar_charts_config["x"] = (
            horizontal_bar_charts_config["x"]
            + model_config["extra_horizontal_bar_charts"]
        )
    if not chart_style:
        chart_style = probe_config["chart_style"]
    if not probe_line_config:
        probe_line_config = model_config["probe_line"]
    if not model_line_config:
        model_line_config = model_config["model_line"]
    if not estimates_selections:
        estimates_selections = model_config["estimates_selections"]
    if not estimates_filters:
        estimates_filters = model_config["estimates_filters"]

    predictor["legend_predictor"] = main_chart_config["legend_title"]
    predictor["legend_error_band"] = error_line_config["legend_title"]
    predictor["legend_model"] = type(fitted_model).__name__
    predictor = filter_predictor(
        predictor=predictor, care_site_level=care_site_level, **kwargs
    )
    for estimate in fitted_model._coefs + fitted_model._metrics:
        if pd.api.types.is_datetime64_any_dtype(predictor[estimate]):
            predictor[estimate] = predictor[estimate].dt.strftime("%Y-%m")

    base = alt.Chart(predictor)
    time_line, time_selection = generate_time_line(
        base=base,
        time_line_config=time_line_config,
    )

    horizontal_bar_charts, y_variables_selections = generate_horizontal_bar_charts(
        base=base,
        horizontal_bar_charts_config=horizontal_bar_charts_config,
        predictor=predictor,
    )
    vertical_bar_charts, x_variables_selections = generate_vertical_bar_charts(
        base=base,
        vertical_bar_charts_config=vertical_bar_charts_config,
        predictor=predictor,
    )

    selections = dict(
        date=time_selection,
        **y_variables_selections,
        **x_variables_selections,
    )
    selection_charts = dict(
        horizontal_bar_charts,
        **vertical_bar_charts,
    )
    base = add_interactive_selection(
        base=base,
        selection_charts=selection_charts,
        selections=selections,
    )
    base = add_estimates_filters(
        base=base,
        selection_charts=selection_charts,
        estimates_filters=estimates_filters,
    )
    index_selection, index_fields = create_groupby_selection(
        indexes=vertical_bar_charts_config["x"] + horizontal_bar_charts_config["y"],
        predictor=predictor,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )
    probe_line = generate_probe_line(
        main_chart=main_chart, probe_line_config=probe_line_config
    )
    error_line = generate_error_line(
        main_chart=main_chart, error_line_config=error_line_config
    )
    model_line = generate_model_line(
        main_chart=main_chart, model_line_config=model_line_config
    )

    main_chart = probe_line + error_line + model_line
    if index_selection:
        main_chart = main_chart.add_params(index_selection)
    chart = concatenate_charts(
        main_chart=main_chart,
        time_line=time_line,
        horizontal_bar_charts=horizontal_bar_charts,
        vertical_bar_charts=vertical_bar_charts,
        spacing=0,
    )
    chart = configure_style(chart=chart, chart_style=chart_style)
    for estimate_selection in estimates_selections:
        chart = chart.add_params(estimate_selection)

    vis_threshold = "id" + uuid.uuid4().hex
    new_sliders_threshold_id = "id" + uuid.uuid4().hex
    old_sliders_threshold_id = "id" + uuid.uuid4().hex
    new_index_threshold_id = "id" + uuid.uuid4().hex
    old_index_threshold_id = "id" + uuid.uuid4().hex
    html_chart = f"""
        <!DOCTYPE html>
        <html>
        <head>
          <script src="https://cdn.jsdelivr.net/npm/vega@{alt.VEGA_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-lite@{alt.VEGALITE_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-embed@{alt.VEGAEMBED_VERSION}"></script>
        </head>
        <body>

        <div class="container">
          <div class="row">
            <div>
            <div id={vis_threshold}></div>
            </div>
            <div style="position:absolute;left:1000px;top:540px;width: -webkit-fill-available;">
            <div id={new_sliders_threshold_id}>
              <div id={old_sliders_threshold_id}></div>
            </div>
            </div>
            <div style="position:absolute;left:45px;top:410px;width: -webkit-fill-available;">
            <div id={new_index_threshold_id}>
              <div id={old_index_threshold_id}></div>
            </div>
            <hr/>
            <h1 style="text-align:center"> Interactive filters </h1>
            </div>
          </div>
        </div>

        <script type="text/javascript">
        vegaEmbed('#{vis_threshold}', {chart.to_json(indent=None)}).then(function(result) {{
            const sliders = document.getElementsByClassName('vega-bindings');
            const newestimate = document.getElementById('{new_sliders_threshold_id}');
            const oldestimate = document.getElementById('{old_sliders_threshold_id}');
            const newparent = document.getElementById('{new_index_threshold_id}');
            const oldchild = document.getElementById('{old_index_threshold_id}');
            for (var i = 0; i < sliders.length; i++) {{
                if (sliders[i].parentElement.parentElement.id == '{vis_threshold}') {{
                    var estimate_slider = sliders[i]
                    var index_slider = estimate_slider.querySelectorAll(".vega-bind")
                    }}
                }}
            newestimate.replaceChild(estimate_slider, oldestimate);
            for (var i = 0; i < index_slider.length; i++) {{
                if (index_slider[i].firstChild.innerHTML == "Group by: ") {{
                    var index_color = index_slider[i]}}
                }}
            newparent.replaceChild(index_color, oldchild);
            }}).catch(console.error);
        </script>
        </body>
        </html>
        """
    display(HTML(html_chart))
    if save_path:
        save_html(
            obj=html_chart,
            filename=save_path,
        )



from .defaults import horizontal_min_c0, model_line, probe_line


def get_probe_dashboard_config(self):
    return dict(
        probe_line=probe_line,
        model_line=model_line,
        extra_horizontal_bar_charts=[horizontal_min_c0],
        extra_vertical_bar_charts=[],
    )



from copy import deepcopy
from datetime import datetime
from typing import Dict, List, Union

import altair as alt
import pandas as pd

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.utils import (
    add_estimates_filters,
    configure_style,
    create_groupby_selection,
    filter_predictor,
    generate_error_line,
    generate_main_chart,
    generate_model_line,
    generate_probe_line,
    month_diff,
    save_html,
)


def normalized_probe_plot(
    probe: BaseProbe,
    fitted_model: BaseModel,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
    care_site_short_name: List[int] = None,
    t_min: int = None,
    t_max: int = None,
    save_path: str = None,
    x_axis_title: str = None,
    y_axis_title: str = None,
    main_chart_config: Dict[str, float] = None,
    model_line_config: Dict[str, str] = None,
    probe_line_config: Dict[str, str] = None,
    error_line_config: Dict[str, str] = None,
    estimates_selections: Dict[str, str] = None,
    estimates_filters: Dict[str, str] = None,
    chart_style: Dict[str, float] = None,
    remove_care_site_id: bool = True,
    **kwargs,
):
    r"""Displays a chart with the aggregated normalized completeness predictor $\frac{c(\Delta t)}{c_0}$ over normalized time $\Delta t = t - t_0$. It represents the overall deviation from the Model.

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$
    fitted_model : BaseModel
        Model fitted to the probe
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    stay_type : List[str], optional
        **EXAMPLE**: `"All"` or `["All", "Urg"]`
    care_site_id : List[int], optional
        **EXAMPLE**: `[8312056386, 8312027648]`
    start_date : datetime, optional
        **EXAMPLE**: `"2019-05-01"`
    end_date : datetime, optional
        **EXAMPLE**: `"2021-07-01"`
    care_site_short_name : List[int], optional
        **EXAMPLE**: `"HOSPITAL XXXX"`
    t_min : int, optional
        Minimal difference with $t_0$ in month $\Delta t_{min}$
        **EXAMPLE**: `-24`
    t_max : int, optional
        Maximal difference with $t_0$ in month $\Delta t_{max}$
        **EXAMPLE**: `24`
    save_path : str, optional
        Folder path where to save the chart in HTML format.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    error_line_config: Dict[str, str], optional
        If not None, configuration used to construct the error line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    estimates_selections: Dict[str, str], optional
        If not None, configuration used to construct the estimates selections.
    estimates_filters: Dict[str, str], optional
        If not None, configuration used to construct the estimates filters.
    chart_style: Dict[str, float], optional
        If not None, configuration used to configure the chart style.
        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    remove_care_site_id: bool, optional
        If False, it will display care site id and care site name, else only care site name.
    """

    predictor = probe.predictor.copy()
    estimates = fitted_model.estimates.copy()

    cols_to_remove = ["date", *probe._metrics]
    if remove_care_site_id:
        cols_to_remove.append("care_site_id")
    indexes = list(set(predictor.columns).difference(cols_to_remove))
    predictor = predictor.merge(estimates, on=probe._index)

    probe_config = deepcopy(probe.get_viz_config("normalized_probe_plot"))
    model_config = deepcopy(
        fitted_model.get_viz_config("normalized_probe_plot", predictor=predictor)
    )
    if not probe_line_config:
        probe_line_config = model_config["probe_line"]
    if not model_line_config:
        model_line_config = model_config["model_line"]
    if not estimates_selections:
        estimates_selections = model_config["estimates_selections"]
    if not estimates_filters:
        estimates_filters = model_config["estimates_filters"]
    if not main_chart_config:
        main_chart_config = probe_config["main_chart"]
    if not error_line_config:
        error_line_config = probe_config["error_line"]
    if not chart_style:
        chart_style = probe_config["chart_style"]

    predictor["normalized_date"] = month_diff(
        predictor["date"], predictor["t_0"]
    ).astype(int)
    predictor["legend_predictor"] = "Mean"
    predictor["legend_error_band"] = "Standard deviation"
    predictor["legend_model"] = type(fitted_model).__name__

    predictor["model"] = 1
    predictor["model"] = predictor["model"].where(predictor["normalized_date"] >= 0, 0)

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
        stay_type=stay_type,
        care_site_id=care_site_id,
        care_site_short_name=care_site_short_name,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )
    for estimate in fitted_model._coefs + fitted_model._metrics:
        if pd.api.types.is_datetime64_any_dtype(predictor[estimate]):
            predictor[estimate] = predictor[estimate].dt.strftime("%Y-%m")

    if t_min:
        predictor = predictor[predictor.normalized_date >= t_min]
    if t_max:
        predictor = predictor[predictor.normalized_date <= t_max]

    indexes = [
        {"field": variable, "title": variable.replace("_", " ").capitalize()}
        for variable in indexes
        if variable in predictor.columns
    ]

    index_selection, index_fields = create_groupby_selection(
        indexes=indexes,
        predictor=predictor,
    )
    base = alt.Chart(predictor)
    base = add_estimates_filters(
        base=base,
        estimates_filters=estimates_filters,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )
    probe_line = generate_probe_line(
        main_chart=main_chart, probe_line_config=probe_line_config
    )
    error_line = generate_error_line(
        main_chart=main_chart, error_line_config=error_line_config
    )
    model_line = generate_model_line(
        main_chart=main_chart, model_line_config=model_line_config
    )
    main_chart = probe_line + error_line + model_line
    if index_selection:
        main_chart = main_chart.add_params(index_selection)

    for estimate_selection in estimates_selections:
        main_chart = main_chart.add_params(estimate_selection)

    main_chart = configure_style(chart=main_chart, chart_style=chart_style)

    if save_path:
        save_html(
            obj=main_chart,
            filename=save_path,
        )

    return main_chart



from typing import Dict

import altair as alt
import pandas as pd

from edsteva.viz.utils import (
    add_interactive_selection,
    concatenate_charts,
    configure_style,
    create_groupby_selection,
    generate_horizontal_bar_charts,
    generate_main_chart,
    generate_model_line,
    generate_probe_line,
    generate_time_line,
    generate_vertical_bar_charts,
)


def fitted_probe_dashboard(
    predictor: pd.DataFrame,
    legend_predictor: str,
    legend_model: str,
    x_axis_title: str,
    y_axis_title: str,
    main_chart_config: Dict[str, float],
    model_line_config: Dict[str, str],
    probe_line_config: Dict[str, str],
    vertical_bar_charts_config: Dict[str, str],
    horizontal_bar_charts_config: Dict[str, str],
    time_line_config: Dict[str, str],
    chart_style: Dict[str, float],
):
    r"""Script to be used by [``predictor_dashboard()``][edsteva.viz.dashboards.probe.wrapper]

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe with its prediction $\hat{c}(t)$
    legend_predictor: str, optional,
        Label name for the predictor legend.
    legend_model: str, optional,
        Label name for the model legend.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        Configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        Configuration used to construct the model line.
    probe_line_config: Dict[str, str], optional
        Configuration used to construct the probe line.
    vertical_bar_charts_config: Dict[str, str], optional
        Configuration used to construct the vertical bar charts.
    horizontal_bar_charts_config: Dict[str, str], optional
        Configuration used to construct the horizontal bar charts.
    time_line_config: Dict[str, str], optional
        Configuration used to construct the time line.
    chart_style: Dict[str, float], optional
        Configuration used to configure the chart style.
        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    """
    predictor["legend_predictor"] = legend_predictor
    predictor["legend_model"] = legend_model

    base = alt.Chart(predictor)
    time_line, time_selection = generate_time_line(
        base=base,
        time_line_config=time_line_config,
    )
    horizontal_bar_charts, y_variables_selections = generate_horizontal_bar_charts(
        base=base,
        horizontal_bar_charts_config=horizontal_bar_charts_config,
        predictor=predictor,
    )
    vertical_bar_charts, x_variables_selections = generate_vertical_bar_charts(
        base=base,
        vertical_bar_charts_config=vertical_bar_charts_config,
        predictor=predictor,
    )

    selections = dict(
        date=time_selection,
        **y_variables_selections,
        **x_variables_selections,
    )
    selection_charts = dict(
        horizontal_bar_charts,
        **vertical_bar_charts,
    )

    base = add_interactive_selection(
        base=base, selection_charts=selection_charts, selections=selections
    )

    index_selection, index_fields = create_groupby_selection(
        indexes=vertical_bar_charts_config["x"] + horizontal_bar_charts_config["y"],
        predictor=predictor,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )

    probe_line = generate_probe_line(
        main_chart=main_chart, probe_line_config=probe_line_config
    )

    model_line = generate_model_line(
        main_chart=main_chart, model_line_config=model_line_config
    )

    main_chart = probe_line + model_line
    if index_selection:
        main_chart = main_chart.add_params(index_selection)
    chart = concatenate_charts(
        main_chart=main_chart,
        time_line=time_line,
        horizontal_bar_charts=horizontal_bar_charts,
        vertical_bar_charts=vertical_bar_charts,
    )
    return configure_style(chart=chart, chart_style=chart_style)



import uuid
from copy import deepcopy
from typing import Dict

import altair as alt
from IPython.display import HTML, display

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.dashboards.probe.fitted_probe import fitted_probe_dashboard
from edsteva.viz.dashboards.probe.probe import probe_only_dashboard
from edsteva.viz.utils import filter_predictor, save_html


def probe_dashboard(
    probe: BaseProbe,
    fitted_model: BaseModel = None,
    care_site_level: str = None,
    save_path: str = None,
    legend_predictor: str = "Predictor c(t)",
    legend_model: str = "Model f(t)",
    x_axis_title: str = None,
    y_axis_title: str = None,
    main_chart_config: Dict[str, float] = None,
    model_line_config: Dict[str, str] = None,
    probe_line_config: Dict[str, str] = None,
    vertical_bar_charts_config: Dict[str, str] = None,
    horizontal_bar_charts_config: Dict[str, str] = None,
    time_line_config: Dict[str, str] = None,
    chart_style: Dict[str, float] = None,
    **kwargs,
):
    r"""Displays an interactive chart with:

    - On the top, the aggregated average completeness predictor $c(t)$ over time $t$ with the fitted model $\hat{c}(t)$ if specified.
    - On the bottom, interactive filters including all the concepts in the [Probe][probe] (such as time, care site, number of visits...etc.)

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$.
    fitted_model : BaseModel, optional
        Model fitted to the probe.
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.
        **EXAMPLE**: `"my_folder/my_file.html"`
    legend_predictor: str, optional,
        Label name for the predictor legend.
    legend_model: str, optional,
        Label name for the model legend.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    vertical_bar_charts_config: Dict[str, str], optional
        If not None, configuration used to construct the vertical bar charts.
    horizontal_bar_charts_config: Dict[str, str], optional
        If not None, configuration used to construct the horizontal bar charts.
    time_line_config: Dict[str, str], optional
        If not None, configuration used to construct the time line.
    chart_style: Dict[str, float], optional
        If not None, configuration used to configure the chart style.
        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    """

    alt.data_transformers.enable("default")
    alt.data_transformers.disable_max_rows()

    probe_config = deepcopy(probe.get_viz_config("probe_dashboard"))
    if fitted_model:
        model_config = deepcopy(fitted_model.get_viz_config("probe_dashboard"))
        if not model_line_config:
            model_line_config = model_config["model_line"]
        if not probe_line_config:
            probe_line_config = model_config["probe_line"]
    if not main_chart_config:
        main_chart_config = probe_config["main_chart"]
    if not time_line_config:
        time_line_config = probe_config["time_line"]
    if not vertical_bar_charts_config:
        vertical_bar_charts_config = probe_config["vertical_bar_charts"]
        if fitted_model:
            vertical_bar_charts_config["y"] = (
                vertical_bar_charts_config["y"]
                + model_config["extra_vertical_bar_charts"]
            )
    if not horizontal_bar_charts_config:
        horizontal_bar_charts_config = probe_config["horizontal_bar_charts"]
        if fitted_model:
            horizontal_bar_charts_config["x"] = (
                horizontal_bar_charts_config["x"]
                + model_config["extra_horizontal_bar_charts"]
            )
    if not chart_style:
        chart_style = probe_config["chart_style"]

    predictor = fitted_model.predict(probe) if fitted_model else probe.predictor.copy()
    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
        **kwargs,
    )

    if fitted_model:
        chart = fitted_probe_dashboard(
            predictor=predictor,
            legend_predictor=legend_predictor,
            legend_model=legend_model,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            main_chart_config=main_chart_config,
            model_line_config=model_line_config,
            probe_line_config=probe_line_config,
            vertical_bar_charts_config=vertical_bar_charts_config,
            horizontal_bar_charts_config=horizontal_bar_charts_config,
            time_line_config=time_line_config,
            chart_style=chart_style,
        )
    else:
        chart = probe_only_dashboard(
            predictor=predictor,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            main_chart_config=main_chart_config,
            vertical_bar_charts_config=vertical_bar_charts_config,
            horizontal_bar_charts_config=horizontal_bar_charts_config,
            time_line_config=time_line_config,
            chart_style=chart_style,
        )

    vis_probe = "id" + uuid.uuid4().hex
    new_index_probe_id = "id" + uuid.uuid4().hex
    old_index_probe_id = "id" + uuid.uuid4().hex
    left_shift = "145px" if fitted_model else "45px"
    html_chart = f"""
        <!DOCTYPE html>
        <html>
        <head>
          <script src="https://cdn.jsdelivr.net/npm/vega@{alt.VEGA_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-lite@{alt.VEGALITE_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-embed@{alt.VEGAEMBED_VERSION}"></script>
        </head>
        <body>

        <div class="container">
          <div class="row">
            <div>
            <div id={vis_probe}></div>
            </div>
            <div style="position:absolute;left:{left_shift};top:380px;width: -webkit-fill-available;">
            <div id={new_index_probe_id}>
              <div id={old_index_probe_id}></div>
            </div>
            <hr/>
            <h1 style="text-align:center"> Interactive filters </h1>
            </div>
          </div>
        </div>

        <script type="text/javascript">
        vegaEmbed('#{vis_probe}', {chart.to_json(indent=None)}).then(function(result) {{
            const sliders = document.getElementsByClassName('vega-bindings');
            const newparent = document.getElementById('{new_index_probe_id}');
            const oldchild = document.getElementById('{old_index_probe_id}');
            for (var i = 0; i < sliders.length; i++) {{
                if (sliders[i].parentElement.parentElement.id == '{vis_probe}') {{
                    var index_slider = sliders[i]
                    }}
                }}
            newparent.replaceChild(index_slider, oldchild);
            }}).catch(console.error);
        </script>
        </body>
        </html>
        """
    if save_path:
        save_html(
            obj=html_chart,
            filename=save_path,
        )
    else:
        display(HTML(html_chart))



import catalogue

from .per_note import compute_completeness_predictor_per_note
from .per_visit import compute_completeness_predictor_per_visit

completeness_predictors = catalogue.create(
    "edsteva.probes.note", "completeness_predictors"
)

completeness_predictors.register(
    "per_visit_default", func=compute_completeness_predictor_per_visit
)
completeness_predictors.register(
    "per_note_default", func=compute_completeness_predictor_per_note
)



from typing import Dict

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession
from pyspark.sql import functions as F
from pyspark.sql.functions import udf as FunctionUDF

from .settings import (
    dict_code_UFR,
    i2b2_renaming,
    i2b2_tables,
    sex_cd_mapping,
    visit_type_mapping,
)


def get_i2b2_table(
    spark_session: SparkSession, db_name: str, db_source: str, table: str
) -> SparkDataFrame:  # pragma: no cover
    """
    Retrieve a Spark table in i2b2 and transform it to fit with OMOP standard.

    Parameters
    ----------
    db_name: str
        Name of the database where the data is stored.
    table: str
        Name of the table to extract.

    Returns
    -------
    df: Spark DataFrame
        Spark DataFrame extracted from the i2b2 database given and converted to OMOP standard.
    """

    table_name = i2b2_tables[db_source][table]
    columns = i2b2_renaming[table]
    if db_source == "cse":
        columns.pop("i2b2_action", None)
    query = ",".join(["{k} AS {v}".format(k=k, v=v) for k, v in columns.items()])

    df = spark_session.sql(f"""SELECT {query} FROM {db_name}.{table_name}""")

    # Special mapping for i2b2 :

    # CIM10
    if table == "condition_occurrence":
        df = df.withColumn(
            "condition_source_value",
            F.split(F.col("condition_source_value"), ":").getItem(1),
        )

        # Retrieve UF
        df = df.withColumn(
            "care_site_source_value",
            F.split(F.col("care_site_source_value"), ":").getItem(1),
        )

    # CCAM
    elif table == "procedure_occurrence":
        df = df.withColumn(
            "procedure_source_value",
            F.split(F.col("procedure_source_value"), ":").getItem(1),
        )

    # Visits
    elif table == "visit_occurrence":
        df = df.withColumn(
            "visit_source_value",
            mapping_dict(visit_type_mapping, "Non Renseigné")(
                F.col("visit_source_value")
            ),
        )
        if db_source == "cse":
            df = df.withColumn("row_status_source_value", F.lit("Actif"))
            df = df.withColumn(
                "visit_occurrence_source_value", df["visit_occurrence_id"]
            )
        else:
            df = df.withColumn(
                "row_status_source_value",
                F.when(
                    F.col("row_status_source_value").isin([-1, -2]), "supprimé"
                ).otherwise("Actif"),
            )
        # Retrieve Hospital trigram
        ufr = spark_session.sql(
            f"SELECT * FROM {db_name}.{i2b2_tables[db_source]['visit_detail']}"
        )
        ufr = ufr.withColumn(
            "care_site_id",
            F.substring(F.split(F.col("concept_cd"), ":").getItem(1), 1, 3),
        )
        ufr = ufr.withColumnRenamed("encounter_num", "visit_occurrence_id")
        ufr = ufr.drop_duplicates(subset=["visit_occurrence_id"])
        ufr = ufr.select(["visit_occurrence_id", "care_site_id"])
        df = df.join(ufr, how="inner", on=["visit_occurrence_id"])

    # Patients
    elif table == "person":
        df = df.withColumn(
            "gender_source_value",
            mapping_dict(sex_cd_mapping, "Non Renseigné")(F.col("gender_source_value")),
        )

    # Documents
    elif table == "note":
        df = df.withColumn(
            "note_class_source_value",
            F.substring(F.col("note_class_source_value"), 4, 100),
        )
        if db_source == "cse":
            df = df.withColumn("row_status_source_value", F.lit("Actif"))
        else:
            df = df.withColumn(
                "row_status_source_value",
                F.when(F.col("row_status_source_value") < 0, "SUPP").otherwise("Actif"),
            )

    # Hospital trigrams
    elif table == "care_site":
        df = df.withColumn("care_site_type_source_value", F.lit("Hôpital"))
        df = df.withColumn(
            "care_site_source_value",
            F.split(F.col("care_site_source_value"), ":").getItem(1),
        )
        df = df.withColumn(
            "care_site_id", F.substring(F.col("care_site_source_value"), 1, 3)
        )
        df = df.drop_duplicates(subset=["care_site_id"])
        df = df.withColumn(
            "care_site_short_name",
            mapping_dict(dict_code_UFR, "Non Renseigné")(F.col("care_site_id")),
        )

    # UFR
    elif table == "visit_detail":
        df = df.withColumn(
            "care_site_id", F.split(F.col("care_site_id"), ":").getItem(1)
        )
        df = df.withColumn("visit_detail_type_source_value", F.lit("PASS"))
        df = df.withColumn("row_status_source_value", F.lit("Actif"))

    # biology
    elif table == "biology":
        df = df.withColumn(
            "biology_source_value", F.substring(F.col("biology_source_value"), 5, 20)
        )

    # fact_relationship
    elif table == "fact_relationship":
        # Retrieve UF information
        df = df.withColumn(
            "fact_id_1",
            F.split(F.col("care_site_source_value"), ":").getItem(1),
        )
        df = df.withColumn("domain_concept_id_1", F.lit(57))  # Care_site domain

        # Retrieve hospital information
        df = df.withColumn("fact_id_2", F.substring(F.col("fact_id_1"), 1, 3))
        df = df.withColumn("domain_concept_id_2", F.lit(57))  # Care_site domain
        df = df.drop_duplicates(subset=["fact_id_1", "fact_id_2"])

        # Only UF-Hospital relationships in i2b2
        df = df.withColumn("relationship_concept_id", F.lit(46233688))  # Included in

    return df


def mapping_dict(
    mapping: Dict[str, str], not_specified_val: str
) -> FunctionUDF:  # pragma: no cover
    """
    Returns a function that maps data according to a mapping dictionnary in a Spark DataFrame.

    Parameters
    ----------
    mapping: Dict
        Mapping dictionnary
    not_specified_val: str
        Value to return if the function input is not find in the mapping dictionnary.

    Returns
    -------
    Callable
        Function that maps the values of Spark DataFrame column.
    """

    def f(x):
        if x in mapping:
            return mapping.get(x)
        return not_specified_val

    return F.udf(f)



import numpy as np
import pandas as pd
from loguru import logger

from edsteva.io.synthetic.utils import (
    generate_events_after_t0,
    generate_events_after_t1,
    generate_events_around_t0,
    generate_events_around_t1,
    generate_events_before_t0,
)


def generate_stays(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    increase_time: int,
    increase_ratio: float,
    care_site_id: int,
    date_col: str,
    mode: str,
):
    if mode == "step":
        return _generate_stays_step(
            generator=generator,
            t_start=t_start,
            t_end=t_end,
            n_events=n_events,
            increase_time=increase_time,
            increase_ratio=increase_ratio,
            care_site_id=care_site_id,
            date_col=date_col,
        )
    if mode == "rect":
        return _generate_stays_rect(
            generator=generator,
            t_start=t_start,
            t_end=t_end,
            n_events=n_events,
            increase_time=increase_time,
            increase_ratio=increase_ratio,
            care_site_id=care_site_id,
            date_col=date_col,
        )


def _generate_stays_step(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    increase_time: int,
    increase_ratio: float,
    care_site_id: int,
    date_col: str,
):
    t0 = generator.integers(t_start + increase_time, t_end - increase_time)
    params = dict(
        generator=generator,
        t_start=t_start,
        t_end=t_end,
        n_events=n_events,
        t0=t0,
        increase_ratio=increase_ratio,
        increase_time=increase_time,
    )
    df = pd.concat(
        [
            generate_events_before_t0(**params),
            generate_events_after_t0(**params),
            generate_events_around_t0(**params),
        ]
    ).to_frame()
    df.columns = [date_col]
    df["care_site_id"] = care_site_id
    df["t_0_min"] = t0 - increase_time / 2
    df["t_0_max"] = t0 + increase_time / 2
    logger.debug("Generate visit occurrences deploying as step function")

    return df


def _generate_stays_rect(
    generator: np.random.Generator,
    t_start: int,
    t_end: int,
    n_events: int,
    increase_time: int,
    increase_ratio: float,
    care_site_id: int,
    date_col: str,
):
    t0 = generator.integers(
        t_start + increase_time, (t_end + t_start) / 2 - increase_time
    )
    t1 = generator.integers(
        (t_end + t_start) / 2 + increase_time, t_end - increase_time
    )
    t0_params = dict(
        generator=generator,
        t_start=t_start,
        t_end=t1 - increase_time / 2,
        n_events=n_events,
        t0=t0,
        increase_ratio=increase_ratio,
        increase_time=increase_time,
    )
    before_t0 = generate_events_before_t0(**t0_params)
    around_t0 = generate_events_around_t0(**t0_params)
    # Raise n_visit to enforce a rectangle shape
    between_t0_t1 = generate_events_after_t0(**t0_params)
    t1_params = dict(
        generator=generator,
        t_start=t_start,
        t_end=t_end,
        n_events=n_events,
        t1=t1,
        increase_time=increase_time,
        increase_ratio=increase_ratio,
    )
    around_t1 = generate_events_around_t1(**t1_params)
    after_t1 = generate_events_after_t1(**t1_params)

    df = pd.concat(
        [
            before_t0,
            around_t0,
            between_t0_t1,
            around_t1,
            after_t1,
        ]
    ).to_frame()

    df.columns = [date_col]
    df["care_site_id"] = care_site_id
    df["t_0_min"] = t0 - increase_time / 2
    df["t_0_max"] = t0 + increase_time / 2
    df["t_1_min"] = t1 - increase_time / 2
    df["t_1_max"] = t1 + increase_time / 2
    logger.debug("Generate visit occurrences deploying as rectangle function")

    return df



import datetime
from abc import ABCMeta, abstractmethod
from typing import ClassVar, Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva import CACHE_DIR
from edsteva.probes.utils.filter_df import filter_table_by_care_site
from edsteva.probes.utils.prepare_df import prepare_care_site_relationship
from edsteva.utils.checks import check_columns, check_tables
from edsteva.utils.file_management import delete_object, load_object, save_object
from edsteva.utils.typing import Data, DataFrame


class BaseProbe(metaclass=ABCMeta):
    """Base class for Probes

    Attributes
    ----------
    _schema: List[str]
        The columns a predictor must have

        **VALUE**: ``["care_site_id", "care_site_level", "stay_type", "date", "c"]``
    predictor: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method
    _cache_predictor: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method

        It is a copy of the predictor DataFrame used to [``reset_predictor()``][edsteva.probes.base.BaseProbe.reset_predictor]
    care_site_relationship: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method

        It describes the care site structure (cf. [``prepare_care_site_relationship()``][edsteva.probes.utils.prepare_df.prepare_care_site_relationship])
    """

    _schema: ClassVar[List[str]] = ["care_site_level", "care_site_id", "date", "c"]

    def __init__(
        self,
        completeness_predictor: str,
        index: List[str],
    ):
        self._completeness_predictor = completeness_predictor
        self._cache_index = index.copy()
        self._viz_config = {}

    def validate_input_data(self, data: Data) -> None:
        """Raises an error if the input data is not valid

        Parameters
        ----------
        data: Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        """

        if not isinstance(data, Data.__args__):
            raise TypeError("Unsupported type {} for data".format(type(data).__name__))

        check_tables(
            data=data,
            required_tables=[
                "visit_occurrence",
                "care_site",
                "fact_relationship",
            ],
        )

    def is_computed_probe(self) -> None:
        """Raises an error if the Probe has not been computed properly"""
        if hasattr(self, "predictor"):
            if not isinstance(self.predictor, pd.DataFrame):
                raise TypeError(
                    "Predictor must be a Pandas DataFrame and not a {}, please review the process method or your arguments".format(
                        type(self.predictor).__name__
                    )
                )
            if self.predictor.empty:
                raise Exception(
                    "Predictor is empty, please review the process method or your arguments"
                )
            check_columns(
                self.predictor,
                required_columns=self._schema,
            )
            if self.predictor.dtypes["date"] != "datetime64[ns]":
                try:
                    self.predictor["date"] = self.predictor["date"].astype(
                        "datetime64[ns]"
                    )
                except Exception as e:
                    raise TypeError(
                        "Predictor column 'date' type is {} and cannot convert to datetime and return the following error: {}. Please review the process method or your arguments".format(
                            self.predictor.dtypes["date"], e
                        )
                    ) from e
        else:
            raise Exception(
                "Predictor has not been computed, please use the compute method as follow: Predictor.compute()"
            )

    def filter_date_per_care_site(self, target_column: str):
        filtered_predictor = self.predictor.copy()
        predictor_activity = self.predictor[self.predictor[target_column] > 0].copy()
        predictor_activity = (
            predictor_activity.groupby("care_site_id")
            .agg({"date": ["min", "max"]})
            .droplevel(axis="columns", level=0)
            .reset_index()
        )
        filtered_predictor = filtered_predictor.merge(
            predictor_activity, on="care_site_id"
        )
        filtered_predictor = filtered_predictor[
            (filtered_predictor["date"] >= filtered_predictor["min"])
            & (filtered_predictor["date"] <= filtered_predictor["max"])
        ].drop(columns=["min", "max"])
        self.predictor = filtered_predictor

    @abstractmethod
    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        start_date: datetime,
        end_date: datetime,
        care_site_levels: List[str],
        stay_types: Union[str, Dict[str, str]],
        care_site_ids: List[int],
        **kwargs,
    ) -> pd.DataFrame:
        """Process the data in order to obtain a predictor table"""

    def compute(
        self,
        data: Data,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        care_site_ids: List[int] = None,
        with_cache: bool = True,
        **kwargs,
    ) -> None:
        """Calls [``compute_process()``][edsteva.probes.base.BaseProbe.compute_process]


        Here are the following computation steps:

        - check if input data is valid with [``validate_input_data()``][edsteva.probes.base.BaseProbe.validate_input_data] method
        - query care site relationship table with [``prepare_care_site_relationship()``][edsteva.probes.utils.prepare_df.prepare_care_site_relationship]
        - compute predictor with [``compute_process()``][edsteva.probes.base.BaseProbe.compute_process] method
        - check if predictor is valid with [``is_computed_probe()``][edsteva.probes.base.BaseProbe.is_computed_probe] method



        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`

        Attributes
        ----------
        Add to the Probe th following attributes:

        - predictor is the target DataFrame
        - _cache_predictor is a copy of the target DataFrame (used to [``reset_predictor()``][edsteva.probes.base.BaseProbe.reset_predictor])
        - care_site_relationship is a DataFrame with the hierarchy of the care site structure

        Examples
        -------

        ```python
        from edsteva.probes import VisitProbe

        visit = VisitProbe()
        visit.compute(
            data,
            stay_types={"All": ".*", "Urg_and_consult": "urgences|consultation"},
            care_site_levels=["Hospital", "Pole", "UF"],
        )
        visit.predictor.head()
        ```

        | care_site_level          | care_site_id | care_site_short_name | stay_type       | date       | n_visit | c     |
        | :----------------------- | :----------- | :------------------- | :-------------- | :--------- | :------ | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | Urg_and_consult | 2019-05-01 | 233.0   | 0.841 |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'All'           | 2021-04-01 | 393.0   | 0.640 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | Urg_and_consult | 2011-03-01 | 204.0   | 0.497 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'All'           | 2018-08-01 | 22.0    | 0.274 |
        | Hôpital                  | 8312022130   | Care site 3          | Urg_and_consult | 2022-02-01 | 9746.0  | 0.769 |


        """
        self.validate_input_data(data=data)
        self._reset_index()
        care_site_relationship = prepare_care_site_relationship(data=data)
        self.start_date = pd.to_datetime(start_date) if start_date else None
        self.end_date = pd.to_datetime(end_date) if end_date else None
        self.predictor = self.compute_process(
            data=data,
            care_site_relationship=care_site_relationship,
            start_date=start_date,
            end_date=end_date,
            care_site_levels=care_site_levels,
            stay_types=stay_types,
            care_site_ids=care_site_ids,
            **kwargs,
        )
        self.is_computed_probe()
        self.care_site_relationship = care_site_relationship
        self.predictor = self.add_names_columns(self.predictor)
        if with_cache:
            self.cache_predictor()

    def reset_predictor(
        self,
    ) -> None:
        """Reset the predictor to its initial state"""
        self.predictor = self._cache_predictor.copy()

    def cache_predictor(
        self,
    ) -> None:
        """Cache the predictor"""
        self._cache_predictor = self.predictor.copy()
        logger.info(
            "Cache the predictor, you can reset the predictor to this state with the method reset_predictor"
        )

    def filter_care_site(
        self,
        care_site_ids: Union[int, List[int]] = None,
        care_site_short_names: Union[str, List[str]] = None,
        care_site_specialties: Union[str, List[str]] = None,
    ) -> None:
        """Filters all the care sites related to the selected care sites.

        Parameters
        ----------
        care_site_ids : Union[int, List[int]], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : Union[str, List[str]], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        """
        self.predictor = filter_table_by_care_site(
            table_to_filter=self.predictor,
            care_site_relationship=self.care_site_relationship,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
            care_site_specialties=care_site_specialties,
        )
        logger.info("Use probe.reset_predictor() to get back the initial predictor")

    def add_names_columns(self, df: DataFrame):
        if hasattr(self, "care_site_relationship") and "care_site_id" in df.columns:
            df = df.merge(
                self.care_site_relationship[
                    ["care_site_id", "care_site_short_name"]
                ].drop_duplicates(),
                on="care_site_id",
                how="left",
            )
        if hasattr(self, "biology_relationship"):
            concept_codes = [
                "{}_concept_code".format(terminology)
                for terminology in self._standard_terminologies
            ]
            concept_names = [
                "{}_concept_name".format(terminology)
                for terminology in self._standard_terminologies
            ]
            if set(concept_codes).issubset(df.columns):
                df = df.merge(
                    self.biology_relationship[
                        concept_codes + concept_names
                    ].drop_duplicates(),
                    on=concept_codes,
                    how="left",
                )
        return df.reset_index(drop=True)

    def load(self, path=None) -> None:
        """Loads a Probe from local

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.load(path=probe_path)
        ```

        """

        path = path or self._get_path()

        loaded_probe = load_object(path)
        self.__dict__ = loaded_probe.__dict__.copy()
        self.path = path

    def save(self, path: str = None, name: str = None) -> bool:
        """Saves computed Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        name : str, optional
            **EXAMPLE**: `"visit_from_BCT"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.compute(data)
        visit.save(path=probe_path)
        ```

        """

        self.is_computed_probe()

        if name:
            self.name = name
        if not path:
            path = self._get_path()

        self.path = path
        save_object(self, path)

    def delete(self, path: str = None):
        """Delete the saved Probe instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        """

        if not path:
            path = self.path

        delete_object(self, path)

    def _get_path(self):
        base_path = CACHE_DIR / "edsteva" / "probes"
        if hasattr(self, "name"):
            filename = f"{self.name.lower()}.pickle"
        else:
            filename = f"{type(self).__name__.lower()}.pickle"
        return base_path / filename

    def _reset_index(
        self,
    ) -> None:
        """Reset the index to its initial state"""
        self._index = self._cache_index.copy()



import altair as alt

vertical_bar_charts = dict(
    x=[
        {
            "title": "Note type",
            "field": "note_type",
            "type": "nominal",
            "sort": "-y",
        },
        {
            "title": "Stay type",
            "field": "stay_type",
            "type": "nominal",
            "sort": "-y",
        },
        {
            "title": "Length of stay",
            "field": "length_of_stay",
            "type": "nominal",
            "sort": "-y",
        },
    ],
    y=[
        dict(
            y=alt.Y(
                "sum(n_note):Q",
                title="Number of discharge summaries",
                axis=alt.Axis(format="s"),
            ),
            tooltip=alt.Tooltip(
                "sum(n_note):Q",
                format=",",
            ),
            sort={
                "field": "n_note",
                "op": "sum",
                "order": "descending",
            },
        ),
    ],
)


horizontal_bar_charts = dict(
    y=[
        {
            "title": "Care site",
            "field": "care_site_short_name",
            "sort": "-x",
        },
        {
            "title": "Care site specialty",
            "field": "care_site_specialty",
            "sort": "-x",
        },
        {
            "title": "Specialties-set",
            "field": "specialties_set",
            "sort": "-x",
        },
        {
            "title": "Care sites-set",
            "field": "care_sites_set",
            "sort": "-x",
        },
    ],
    x=[
        dict(
            x=alt.X(
                "sum(n_note):Q",
                title="Number of discharge summaries",
                axis=alt.Axis(format="s"),
            ),
            tooltip=alt.Tooltip(
                "sum(n_note):Q",
                format=",",
            ),
            sort={
                "field": "n_note",
                "op": "sum",
                "order": "descending",
            },
        ),
    ],
)

normalized_main_chart = dict(
    aggregates=[
        dict(
            sum_note="sum(n_note)",
            groupby=["value", "date"],
        ),
        dict(
            max_note="max(sum_note)",
            groupby=["value"],
        ),
    ],
    calculates=[
        dict(normalized_c=(alt.datum.sum_note / alt.datum.max_note) / alt.datum.c_0)
    ],
    legend_title="Mean",
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "mean(normalized_c):Q",
            title="c(Δt) / c₀",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            sort={
                "field": "n_note",
                "op": "sum",
                "order": "descending",
            },
            title=None,
        ),
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

main_chart = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "sum(n_note):Q",
            title="Number of discharge summaries",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            sort={
                "field": "n_note",
                "op": "sum",
                "order": "descending",
            },
            title=None,
        ),
        tooltip=[
            alt.Tooltip("value:N", title="Index"),
            alt.Tooltip("yearmonth(date):T", title="Date"),
            alt.Tooltip(
                "sum(n_note):Q", title="Number of discharge summaries", format=","
            ),
        ],
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

normalized_time_line = dict(
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "sum(n_note):Q",
            title="Number of discharge summaries",
            axis=alt.Axis(format="s"),
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

time_line = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "sum(n_note):Q",
            title="Number of discharge summaries",
            axis=alt.Axis(format="s"),
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

error_line = dict(
    legend_title="Standard deviation",
    mark_errorband=dict(
        extent="stdev",
    ),
    encode=dict(
        stroke=alt.Stroke(
            "legend_error_band",
            title="Error band",
            legend=alt.Legend(
                symbolType="square",
                orient="top",
                labelFontSize=12,
                labelFontStyle="bold",
            ),
        ),
    ),
)

chart_style = dict(
    labelFontSize=12,
    titleFontSize=13,
)



import pandas as pd


def generate_care_site_tables(structure, parent=None, final=True):
    cs = []
    fr = []
    for key, value in structure.items():
        this_cs = _split_name_id(key)
        cs.append(this_cs)

        if parent is not None:
            this_fr = dict(
                fact_id_1=this_cs["care_site_id"],
                fact_id_2=parent["care_site_id"],
            )
            fr.append(this_fr)

        if value is not None:
            next_cs, next_fr = generate_care_site_tables(
                value, parent=this_cs, final=False
            )
            cs.extend(next_cs)
            fr.extend(next_fr)

    if final:
        cs = pd.DataFrame(cs)
        fr = pd.DataFrame(fr)
        fr["domain_concept_id_1"] = 57
        fr["relationship_concept_id"] = 46233688

    return cs, fr


def _split_name_id(string):
    splitted = string.split("-")

    # Name - Type - ID
    return dict(
        care_site_short_name=string,
        care_site_type_source_value=splitted[0],
        care_site_id=splitted[1],
    )



from .defaults import chart_style, get_horizontal_bar_charts, vertical_bar_charts


def get_estimates_densities_plot_config(self):
    horizontal_bar_charts = get_horizontal_bar_charts(
        standard_terminologies=self._standard_terminologies.copy()
    )
    return dict(
        chart_style=chart_style,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )



import catalogue

from edsteva.models.step_function.algos.loss_minimization import loss_minimization
from edsteva.models.step_function.algos.quantile import quantile

algos = catalogue.create("edsteva.models.step_function", "algos")

algos.register("loss_minimization", func=loss_minimization)
algos.register("quantile", func=quantile)



import altair as alt

from edsteva.utils.typing import DataFrame
from edsteva.viz.utils import round_up, scale_it


def get_c_0_min_selection(predictor: DataFrame):
    c_0_min_slider = alt.binding_range(
        min=0,
        max=round_up(predictor.c_0.max(), 2),
        step=scale_it(predictor.c_0.max()) / 100,
        name="c₀ min: ",
    )
    c_0_min_selection = alt.selection_point(
        name="c_0_min",
        fields=["c_0_min"],
        bind=c_0_min_slider,
        value=0,
    )
    c_0_min_filter = alt.datum.c_0 >= c_0_min_selection.c_0_min
    return c_0_min_selection, c_0_min_filter


def get_error_max_selection(predictor: DataFrame):
    error_max_slider = alt.binding_range(
        min=0,
        max=round_up(predictor.error.max(), 2),
        step=scale_it(predictor.error.max()) / 100,
        name="error max: ",
    )
    error_max_selection = alt.selection_point(
        name="error_max",
        fields=["error_max"],
        bind=error_max_slider,
        value=round_up(predictor.error.max(), 2),
    )
    error_max_filter = alt.datum.error <= error_max_selection.error_max
    return error_max_selection, error_max_filter


def get_t_0_selection(predictor: DataFrame):
    t_0_slider = alt.binding(
        input="t_0",
        name="t₀ max: ",
    )
    t_0_selection = alt.selection_point(
        name="t_0",
        fields=["t_0"],
        bind=t_0_slider,
        value=predictor.t_0.astype(str).max(),
    )
    t_0_min_filter = alt.datum.t_0 <= t_0_selection.t_0
    return t_0_selection, t_0_min_filter


normalized_probe_line = dict(
    encode=dict(
        strokeDash=alt.StrokeDash(
            "legend_predictor",
            title="Predictor line",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=12,
                labelFontStyle="bold",
                orient="top",
            ),
        )
    )
)
probe_line = dict(
    encode=dict(
        strokeDash=alt.StrokeDash(
            "legend_predictor",
            title="",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=12,
                labelFontStyle="bold",
                orient="left",
            ),
        )
    )
)
normalized_model_line = dict(
    mark_line=dict(
        color="black",
        interpolate="step-after",
        strokeDash=[5, 5],
    ),
    encode=dict(
        y="model:Q",
        strokeWidth=alt.StrokeWidth(
            field="legend_model",
            title="Model line",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=12,
                labelFontStyle="bold",
                symbolDash=[2, 2],
                orient="top",
            ),
        ),
    ),
)

model_line = dict(
    mark_line=dict(
        interpolate="step-after",
        strokeDash=[5, 5],
    ),
    encode=dict(
        y=alt.Y("min(c_hat)"),
        strokeWidth=alt.StrokeWidth(
            field="legend_model",
            title="",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=12,
                labelFontStyle="bold",
                symbolDash=[2, 2],
                orient="left",
            ),
        ),
        tooltip=[
            alt.Tooltip("value:N", title="Index"),
            alt.Tooltip("yearmonth(date):T", title="Date"),
            alt.Tooltip("min(c_hat):Q", title="c₀"),
        ],
    ),
    aggregates=[
        dict(
            max_t0="max(t_0)",
            groupby=["value", "date"],
        ),
    ],
    filters=[dict(filter=alt.datum.t_0 == alt.datum.max_t0)],
)

horizontal_min_c0 = dict(
    x=alt.X(
        "min(c_0):Q",
        title="Min(c₀)",
    ),
    tooltip=alt.Tooltip("min(c_0):Q", format=".2"),
    sort={
        "field": "c_0",
        "op": "min",
        "order": "descending",
    },
)

horizontal_max_error = dict(
    x=alt.X(
        "max(error):Q",
        title="Max(error)",
    ),
    tooltip=alt.Tooltip("max(error):Q", format=".2"),
    sort={
        "field": "error",
        "op": "max",
        "order": "descending",
    },
)



from typing import Dict, List

import altair as alt
import pandas as pd

from edsteva.viz.utils import (
    create_groupby_selection,
    generate_main_chart,
    generate_model_line,
    generate_probe_line,
)


def fitted_probe_line(
    predictor: pd.DataFrame,
    indexes: List[Dict[str, str]],
    legend_predictor: str,
    legend_model: str,
    x_axis_title: str,
    y_axis_title: str,
    main_chart_config: Dict[str, float],
    model_line_config: Dict[str, str],
    probe_line_config: Dict[str, str],
):
    r"""Script to be used by [``plot_probe()``][edsteva.viz.plots.probe.wrapper]

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe with its prediction $\hat{c}(t)$
    indexes : List[str]
        Variable from which data is grouped
    legend_predictor: str, optional,
        Label name for the predictor legend.
    legend_model: str, optional,
        Label name for the model legend.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    """
    predictor["legend_predictor"] = legend_predictor
    predictor["legend_model"] = legend_model

    base = alt.Chart(predictor)

    index_selection, index_fields = create_groupby_selection(
        indexes=indexes,
        predictor=predictor,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )

    probe_line = generate_probe_line(
        main_chart=main_chart, probe_line_config=probe_line_config
    )

    model_line = generate_model_line(
        main_chart=main_chart, model_line_config=model_line_config
    )

    main_chart = probe_line + model_line
    if index_selection:
        main_chart = main_chart.add_params(index_selection)

    return main_chart



from .defaults import chart_style, horizontal_bar_charts, vertical_bar_charts


def get_estimates_densities_plot_config(self):
    return dict(
        chart_style=chart_style,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd

from edsteva.probes.base import BaseProbe
from edsteva.probes.condition.completeness_predictors import completeness_predictors
from edsteva.probes.condition.viz_configs import viz_configs
from edsteva.utils.typing import Data


class ConditionProbe(BaseProbe):
    r"""
    The [``ConditionProbe``][edsteva.probes.condition.condition.ConditionProbe] computes $c(t)$ the availability of claim data:

    Parameters
    ----------
    completeness_predictor: str
        Algorithm used to compute the completeness predictor
        **EXAMPLE**: ``"per_visit_default"``

    Attributes
    ----------
    _completeness_predictor: str
        Algorithm used to compute the completeness predictor
        **VALUE**: ``"per_visit_default"``
    _index: List[str]
        Variable from which data is grouped
        **VALUE**: ``["care_site_level", "stay_type", "length_of_stay", "care_site_specialty", "specialties_set", "diag_type", "condition_type", "source_system", "care_site_id"]``
    _viz_config: List[str]
        Dictionary of configuration for visualization purpose.
        **VALUE**: ``{}``
    """

    def __init__(
        self,
        completeness_predictor: str = "per_visit_default",
    ):
        self._index = [
            "care_site_level",
            "stay_type",
            "length_of_stay",
            "diag_type",
            "condition_type",
            "source_system",
            "care_site_id",
            "care_site_specialty",
            "care_sites_set",
            "specialties_set",
        ]
        super().__init__(
            completeness_predictor=completeness_predictor,
            index=self._index,
        )

    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        start_date: datetime,
        end_date: datetime,
        care_site_levels: List[str],
        stay_types: Union[str, Dict[str, str]],
        care_site_ids: List[int],
        extra_data: Data = None,
        care_site_short_names: List[str] = None,
        care_site_specialties: List[str] = None,
        diag_types: Union[str, Dict[str, str]] = None,
        care_sites_sets: Union[str, Dict[str, str]] = None,
        specialties_sets: Union[str, Dict[str, str]] = None,
        condition_types: Union[str, Dict[str, str]] = None,
        source_systems: List[str] = ["ORBIS"],
        length_of_stays: List[float] = None,
        **kwargs,
    ):
        """Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        care_site_relationship : pd.DataFrame
            DataFrame computed in the [``compute()``][edsteva.probes.base.BaseProbe.compute] that gives the hierarchy of the care site structure.
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés"`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : List[str], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        care_site_specialties : List[str], optional
            **EXAMPLE**: `["CARDIOLOGIE", "CHIRURGIE"]`
        diag_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "DP\DR": "DP|DR"}` or `"DP"`
        care_sites_sets : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All AP-HP": ".*"}` or `{"All AP-HP": ".*", "Pediatrics": r"debre|trousseau|necker"}`
        specialties_sets : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "ICU": r"REA\s|USI\s|SC\s"}`
        condition_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Pulmonary_embolism": "I26"}`
        source_systems : List[str], optional
            **EXAMPLE**: `["AREM", "ORBIS"]`
        length_of_stays : List[float], optional
            **EXAMPLE**: `[1, 30]`
        """
        if specialties_sets is None and "specialties_set" in self._index:
            self._index.remove("specialties_set")
        if length_of_stays is None and "length_of_stay" in self._index:
            self._index.remove("length_of_stay")
        if care_sites_sets is None and "care_sites_set" in self._index:
            self._index.remove("care_sites_set")
        if condition_types is None and "condition_type" in self._index:
            self._index.remove("condition_type")
        return completeness_predictors.get(self._completeness_predictor)(
            self,
            data=data,
            care_site_relationship=care_site_relationship,
            start_date=start_date,
            end_date=end_date,
            care_site_levels=care_site_levels,
            stay_types=stay_types,
            care_site_ids=care_site_ids,
            extra_data=extra_data,
            care_site_short_names=care_site_short_names,
            care_site_specialties=care_site_specialties,
            care_sites_sets=care_sites_sets,
            specialties_sets=specialties_sets,
            diag_types=diag_types,
            length_of_stays=length_of_stays,
            condition_types=condition_types,
            source_systems=source_systems,
            **kwargs,
        )

    def get_viz_config(self, viz_type: str, **kwargs):
        if viz_type in viz_configs.keys():
            _viz_config = self._viz_config.get(viz_type)
            if _viz_config is None:
                _viz_config = self._completeness_predictor
        else:
            raise ValueError(f"edsteva has no {viz_type} registry !")
        return viz_configs[viz_type].get(_viz_config)(self, **kwargs)

    def available_completeness_predictors(self):
        return list(completeness_predictors.get_all().keys())



import catalogue

from edsteva.models.rectangle_function.algos.loss_minimization import loss_minimization

algos = catalogue.create("edsteva.models.rectangle_function", "algos")

algos.register("loss_minimization", func=loss_minimization)



import catalogue

from edsteva.models.step_function.viz_configs.normalized_probe_dashboard import (
    get_normalized_probe_dashboard_config,
)
from edsteva.models.step_function.viz_configs.normalized_probe_plot import (
    get_normalized_probe_plot_config,
)
from edsteva.models.step_function.viz_configs.probe_dashboard import (
    get_probe_dashboard_config,
)
from edsteva.models.step_function.viz_configs.probe_plot import get_probe_plot_config

normalized_probe_dashboard = catalogue.create(
    "edsteva.models.step_function.viz_configs", "normalized_probe_dashboard"
)
normalized_probe_dashboard.register(
    "default", func=get_normalized_probe_dashboard_config
)


probe_dashboard = catalogue.create(
    "edsteva.models.step_function.viz_configs", "probe_dashboard"
)
probe_dashboard.register("default", func=get_probe_dashboard_config)


normalized_probe_plot = catalogue.create(
    "edsteva.models.step_function.viz_configs", "normalized_probe_plot"
)
normalized_probe_plot.register("default", func=get_normalized_probe_plot_config)


probe_plot = catalogue.create("edsteva.models.step_function.viz_configs", "probe_plot")
probe_plot.register("default", func=get_probe_plot_config)

viz_configs = dict(
    normalized_probe_dashboard=normalized_probe_dashboard,
    probe_dashboard=probe_dashboard,
    normalized_probe_plot=normalized_probe_plot,
    probe_plot=probe_plot,
)



import os
from types import ModuleType
from typing import Optional

import pandas as _pandas
import pyarrow.parquet as pq
from databricks import koalas as _koalas
from loguru import logger

from edsteva.utils.typing import DataObject

VALID_FRAMEWORKS = {
    "pandas": _pandas,
    "koalas": _koalas,
}


def get_framework(obj: DataObject) -> Optional[ModuleType]:
    for _, framework in VALID_FRAMEWORKS.items():
        if obj.__class__.__module__.startswith(framework.__name__):
            return framework
    return None


def is_pandas(obj: DataObject) -> bool:
    return get_framework(obj) == _pandas


def is_koalas(obj: DataObject) -> bool:
    return get_framework(obj) == _koalas


def to(framework: str, obj: DataObject) -> DataObject:
    if framework == "koalas" or framework is _koalas:
        return koalas(obj)
    if framework == "pandas" or framework is _pandas:
        return pandas(obj)
    raise ValueError(f"Unknown framework: {framework}")


def pandas(obj: DataObject) -> DataObject:
    if get_framework(obj) is _pandas:
        return obj

    # Try using pyarrow via HDFS to convert object to pandas as it is way faster.
    user = os.environ["USER"]
    parquet_path = f"hdfs://bbsedsi/user/{user}/temp.parquet"
    try:  # pragma: no cover
        obj.to_parquet(parquet_path)
        obj = pq.read_table(parquet_path)
    except Exception as e:
        logger.debug(
            "Cannot convert object to parquet. It will skip this step and convert directly to pandas if possible. /n Following error: {}",
            e,
        )

    try:
        return obj.to_pandas()
    except AttributeError:
        pass
    raise ValueError("Could not convert object to pandas.")


def koalas(obj: DataObject) -> DataObject:
    if get_framework(obj) is _koalas:
        return obj
    try:
        return obj.to_koalas()
    except AttributeError:
        pass

    # will raise ValueError if impossible
    return _koalas.from_pandas(obj)



from edsteva.probes.base import BaseProbe
from edsteva.probes.biology.biology import BiologyProbe
from edsteva.probes.condition.condition import ConditionProbe
from edsteva.probes.note.note import NoteProbe
from edsteva.probes.visit.visit import VisitProbe



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.filter_df import convert_uf_to_pole
from edsteva.probes.utils.prepare_df import (
    prepare_care_site,
    prepare_condition_occurrence,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_condition_source_systems, check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_condition(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: List[str],
    stay_types: Union[str, Dict[str, str]],
    care_site_ids: List[int],
    extra_data: Data,
    care_site_short_names: List[str],
    care_site_specialties: List[str],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    diag_types: Union[str, Dict[str, str]],
    condition_types: Union[str, Dict[str, str]],
    source_systems: List[str],
    length_of_stays: List[float],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_condition`` algorithm computes $c_(t)$ the availability of claim data as follow:

    $$
    c(t) = \frac{n_{condition}(t)}{n_{max}}
    $$

    Where $n_{condition}(t)$ is the number of claim codes (e.g. ICD-10) recorded, $t$ is the month and $n_{max} = \max_{t}(n_{condition}(t))$.
    """

    self._metrics = ["c", "n_condition"]
    check_tables(data=data, required_tables=["condition_occurrence"])
    check_condition_source_systems(source_systems=source_systems)
    if "AREM" in source_systems and not hospital_only(
        care_site_levels=care_site_levels
    ):  # pragma: no cover
        logger.info("AREM claim data are only available at hospital level")

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
    ).drop(columns="date")

    condition_occurrence = prepare_condition_occurrence(
        data=data,
        extra_data=extra_data,
        visit_occurrence=visit_occurrence,
        source_systems=source_systems,
        diag_types=diag_types,
        condition_types=condition_types,
        start_date=start_date,
        end_date=end_date,
    )

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_relationship=care_site_relationship,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
    )

    hospital_visit = get_hospital_condition(
        condition_occurrence,
        visit_occurrence,
        care_site,
    )
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    condition_predictor_by_level = {hospital_name: hospital_visit}

    # UF selection
    if not hospital_only(care_site_levels=care_site_levels):
        visit_detail = prepare_visit_detail(
            data=data,
            start_date=start_date,
            end_date=end_date,
        )

        uf_condition = get_uf_condition(
            condition_occurrence=condition_occurrence,
            visit_occurrence=visit_occurrence,
            visit_detail=visit_detail,
            care_site=care_site,
        )
        uf_name = CARE_SITE_LEVEL_NAMES["UF"]
        condition_predictor_by_level[uf_name] = uf_condition

        pole_condition = get_pole_condition(
            uf_condition, care_site, care_site_relationship
        )
        pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
        condition_predictor_by_level[pole_name] = pole_condition

    condition_predictor = concatenate_predictor_by_level(
        predictor_by_level=condition_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    return compute_completeness(self, condition_predictor)


def compute_completeness(
    self,
    condition_predictor: DataFrame,
):
    partition_cols = [*self._index.copy(), "date"]

    n_condition = (
        condition_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"condition_occurrence_id": "nunique"})
        .rename(columns={"condition_occurrence_id": "n_condition"})
    )
    n_condition = to("pandas", n_condition)
    n_condition = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_condition,
        partition_cols=partition_cols,
    )

    partition_cols = list(set(partition_cols) - {"date"})
    max_n_condition = (
        n_condition.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"n_condition": "max"})
        .rename(columns={"n_condition": "max_n_condition"})
    )

    condition_predictor = n_condition.merge(
        max_n_condition,
        on=partition_cols,
    )

    condition_predictor["c"] = condition_predictor["max_n_condition"].where(
        condition_predictor["max_n_condition"] == 0,
        condition_predictor["n_condition"] / condition_predictor["max_n_condition"],
    )
    return condition_predictor.drop(columns="max_n_condition")


def get_hospital_condition(
    condition_occurrence: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    hospital_condition = condition_occurrence.merge(
        visit_occurrence,
        on="visit_occurrence_id",
    )
    hospital_condition = hospital_condition.drop(columns="visit_occurrence_id")
    hospital_condition = hospital_condition.merge(care_site, on="care_site_id")
    if is_koalas(hospital_condition):
        hospital_condition = hospital_condition.spark.cache()

    return hospital_condition


def get_uf_condition(
    condition_occurrence: DataFrame,
    visit_occurrence: DataFrame,
    visit_detail: DataFrame,
    care_site: DataFrame,
):
    # Add visit information
    visit_detail = visit_detail[visit_detail.visit_detail_type == "RUM"]
    uf_condition = condition_occurrence.merge(
        visit_occurrence.drop(columns="care_site_id"),
        on="visit_occurrence_id",
    ).rename(columns={"visit_detail_id": "visit_id"})

    # Add care_site information
    uf_condition = uf_condition.merge(
        visit_detail[["visit_id", "care_site_id"]],
        on="visit_id",
        how="left",
    )
    uf_condition = uf_condition.merge(care_site, on="care_site_id")

    uf_name = CARE_SITE_LEVEL_NAMES["UF"]
    uf_condition = uf_condition[uf_condition["care_site_level"] == uf_name]

    if is_koalas(uf_condition):
        uf_condition = uf_condition.spark.cache()

    return uf_condition


def get_pole_condition(
    uf_condition: DataFrame,
    care_site: DataFrame,
    care_site_relationship: DataFrame,
):
    care_site_cols = list(
        set(
            [
                "care_site_short_name",
                "care_site_level",
                "care_site_specialty",
                "specialties_set",
                "care_sites_set",
            ]
        ).intersection(uf_condition.columns)
    )
    pole_condition = convert_uf_to_pole(
        table=uf_condition.drop(columns=care_site_cols),
        table_name="uf_condition",
        care_site_relationship=care_site_relationship,
    )

    pole_condition = pole_condition.merge(care_site, on="care_site_id")

    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    pole_condition = pole_condition[pole_condition["care_site_level"] == pole_name]

    if is_koalas(pole_condition):
        pole_condition = pole_condition.spark.cache()

    return pole_condition



from copy import deepcopy
from datetime import datetime
from typing import Dict, List

import altair as alt

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.plots.probe.fitted_probe import fitted_probe_line
from edsteva.viz.plots.probe.probe import probe_line
from edsteva.viz.utils import configure_style, filter_predictor, save_html


def probe_plot(
    probe: BaseProbe,
    fitted_model: BaseModel = None,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    start_date: datetime = None,
    end_date: datetime = None,
    care_site_short_name: List[int] = None,
    save_path: str = None,
    legend_predictor: str = "Predictor c(t)",
    legend_model: str = "Model f(t)",
    x_axis_title: str = None,
    y_axis_title: str = None,
    main_chart_config: Dict[str, float] = None,
    model_line_config: Dict[str, str] = None,
    probe_line_config: Dict[str, str] = None,
    chart_style: Dict[str, float] = None,
    remove_care_site_id: bool = True,
    **kwargs,
):
    r"""
    Displays a chart with the average completeness predictor $c(t)$ over time $t$ with the fitted model $\hat{c}(t)$ if specified.
    The chart is exportable in png or svg format and easy to integrate into a report. Is also possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$.
    fitted_model : BaseModel, optional
        Model fitted to the probe
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    stay_type : List[str], optional
        **EXAMPLE**: `"All"` or `["All", "Urg"]`
    care_site_id : List[int], optional
        **EXAMPLE**: `[8312056386, 8312027648]`
    start_date : datetime, optional
        **EXAMPLE**: `"2019-05-01"`
    end_date : datetime, optional
        **EXAMPLE**: `"2021-07-01"`
    care_site_short_name : List[int], optional
        **EXAMPLE**: `"HOSPITAL XXXX"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.
        **EXAMPLE**: `"my_folder/my_file.html"`
    legend_predictor: str, optional,
        Label name for the predictor legend.
    legend_model: str, optional,
        Label name for the model legend.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    chart_style: Dict[str, float], optional
        If not None, configuration used to configure the chart style.
        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    remove_care_site_id: bool, optional
        If False, it will display care site id and care site name, else only care site name.
    """
    alt.data_transformers.enable("default")
    alt.data_transformers.disable_max_rows()

    probe_config = deepcopy(probe.get_viz_config("probe_plot"))
    if not main_chart_config:
        main_chart_config = probe_config["main_chart"]
    if not chart_style:
        chart_style = probe_config["chart_style"]
    predictor = probe.predictor.copy()
    cols_to_remove = ["date", *probe._metrics]
    if remove_care_site_id:
        cols_to_remove.append("care_site_id")
    indexes = list(set(predictor.columns).difference(cols_to_remove))

    if fitted_model:
        predictor = fitted_model.predict(probe).copy()
    else:
        predictor = probe.predictor.copy()

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
        stay_type=stay_type,
        care_site_id=care_site_id,
        care_site_short_name=care_site_short_name,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )

    indexes = [
        {"field": variable, "title": variable.replace("_", " ").capitalize()}
        for variable in indexes
        if variable in predictor.columns and len(predictor[variable].unique()) >= 2
    ]

    if fitted_model:
        model_config = deepcopy(fitted_model.get_viz_config("probe_plot"))
        if not model_line_config:
            model_line_config = model_config["model_line"]
        if not probe_line_config:
            probe_line_config = model_config["probe_line"]
        chart = fitted_probe_line(
            predictor=predictor,
            indexes=indexes,
            legend_predictor=legend_predictor,
            legend_model=legend_model,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            main_chart_config=main_chart_config,
            model_line_config=model_line_config,
            probe_line_config=probe_line_config,
        )
    else:
        chart = probe_line(
            predictor=predictor,
            indexes=indexes,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            main_chart_config=main_chart_config,
        )

    if save_path:
        save_html(
            obj=configure_style(chart=chart, chart_style=chart_style),
            filename=save_path,
        )

    return chart
